import functools
import inspect

import pandas as pd
import numpy as np

from pyadlml.constants import TIME, END_TIME, START_TIME, DEVICE, VALUE, BOOL, CAT, ACTIVITY
from pandas.api.types import infer_dtype
from pyadlml.dataset._core.activities import _is_activity_overlapping, correct_activity_overlap

"""
    includes generic methods for manpulating dataframes
"""

def print_df(df):
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        print(df)

def unitsfromdaystart(ts, unit='s'):
    """ Computes units passed from the start of the day until the timestamp
    """
    ts = pd.Timestamp(ts) if not isinstance(ts, pd.Timestamp) else ts
    seconds = ts.hour*3600 + ts.minute*60 + ts.second
    if unit == 's':
        return int(seconds)
    elif unit == 'm':
        return int(seconds/60)
    elif unit == 'h':
        return int(seconds/3600)
    else:
        raise ValueError





def timestr_2_timedeltas(t_strs):
    """
        gets either a string or a list of strings to convert to a list of 
        timedeltas
    """
    if isinstance(t_strs, list):
        return [timestr_2_timedelta(t_str) for t_str in t_strs]
    else:
        return [timestr_2_timedelta(t_strs)]


def timestr_2_timedelta(t_str):
    """
        t_str (string)
        of form 30s, 30m
    """
    ttype = t_str[-1:]
    val = int(t_str[:-1])

    assert ttype in ['h','m','s']
    assert (val > 0 and val <=12 and ttype == 'h')\
        or (val > 0 and val <= 60 and ttype == 'm')\
        or (val > 0 and val <= 60 and ttype == 's')
    import datetime as dt
    if ttype == 's':
        return pd.Timedelta(seconds=val)
    if ttype == 'm':
        return pd.Timedelta(seconds=val*60)
    if ttype == 'h':
        return pd.Timedelta(seconds=val*3600)


def time2int(ts, t_res='30m'):
    """
    rounds to the next lower min bin or hour bin
    """
    assert t_res[-1:] in ['h','m']
    val = int(t_res[:-1])

    assert (val > 0 and val <=12 and t_res[-1:] == 'h')\
        or (val > 0 and val <= 60 and t_res[-1:] == 'm')

    import datetime as dt
    zero = dt.time()

    if t_res[-1:] == 'h':
        hs = val
        h_bin = int(ts.hour/hs)*hs
        return dt.time(hour=h_bin)

    elif t_res[-1:] == 'm':
        ms = val
        m_bin = int(ts.minute/ms)*ms
        return dt.time(hour=ts.hour, minute=m_bin)
    else:
        raise ValueError



def fill_nans_ny_inverting_first_occurence(df):
    """
    fills up true or false values
    :param df:
                Name           0      1         10
        time                                     
        2008-02-25 00:20:14    NaN    NaN   ...    True
        2008-02-25 00:22:57    NaN    NaN   ...    False
        2008-02-25 09:33:41    NaN    True  ...    True
        2008-02-25 09:33:42    False    NaN   ...    False
    :return:
                Name           0      1        10   
        Time                                       
        2008-02-25 00:20:14    True   False ... True 
        2008-02-25 00:22:57    True   False ... False 
        2008-02-25 09:33:41    True   True  ... True 
        2008-02-25 09:33:42    False   NaN  ... False 
    """
    for col_label in df.columns:
        col = df[col_label]

        # get timestamp of first valid index and replace previous Nans 
        #   by opposite
        ts = col.first_valid_index()
        idx = df.index.get_loc(ts)
        col.iloc[0:idx] = not col[ts]
    return df


def categorical_2_binary(df_devices, cat_list):
    """
    Transforms all categorical devices within the device dataframe into
    binary by creating a new device 'cat:cat_value' that reports a 1
    when the category is activated and 0 when it ends

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe
    cat_list : lst
        Categorical devices from the dataframe

    Returns
    -------
    res : pd.DataFrame
        A device dataframe with newly created devices

    """
    mask_cat = df_devices[DEVICE].isin(cat_list)
    df_cat = df_devices[mask_cat].copy()
    df_cat.loc[:, 'new_device'] = df_cat[DEVICE] + ':' + df_cat[VALUE]
    df_cat.loc[:, VALUE] = True
    for dev in df_cat[DEVICE].unique():
        df = df_cat[df_cat[DEVICE] == dev]
        df_new = df.copy()
        df_new.loc[:, DEVICE] = df['new_device'].shift(1)
        df_new = df_new[1:]
        df_new.loc[:, VALUE] = False
        df_new.loc[:, TIME] += pd.Timedelta('1ns')
        df_cat = pd.concat([df_cat, df_new])

    mask = df_cat[DEVICE].isin(cat_list)
    df_cat.loc[mask, DEVICE] = df_cat.loc[mask, 'new_device']
    df_cat = df_cat.drop(columns='new_device')

    df_devices = pd.concat([df_devices[~mask_cat], df_cat])\
                    .sort_values(by=TIME)\
                    .reset_index(drop=True)
    return df_devices

def infer_dtypes(df_devices):
    """ Infers automatically the datatypes for each device of a device dataframe
    and returns a dictionary containing data types mapped to device names

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe

    Returns
    -------
    res : dict
        Dictionary with keys: {'categorical' : [...], 'boolean': [...] , 'numerical' : [...]}

    """
    dev_cat = []
    dev_bool = []
    dev_num = []

    dev_lst = df_devices[DEVICE].unique()
    for dev in dev_lst:
        vals = df_devices[df_devices[DEVICE] == dev][VALUE]
        inf = infer_dtype(vals, skipna=True)
        if inf == 'string' or inf == 'object':
            try:
                pd.to_numeric(vals.dropna().unique())
                dev_num.append(dev)
            except:
                dev_cat.append(dev)
        elif inf == 'boolean':
            dev_bool.append(dev)
        elif inf == 'floating':
            dev_num.append(dev)
        else:
            raise ValueError('could not infer correct dtype for device {}'.format(dev))

    return {'categorical': dev_cat, 'boolean': dev_bool, 'numerical': dev_num}


def select_timespan(df_devices=None, df_activities=None, start_time=None, end_time=None, clip_activities=False):
    """ Selects a subset of a device and an activity dataframe based on a time span given
    Parameters
    ----------
    df_devices : pd.DataFrame or None
        A device dataframe
    df_activities : pd.DataFrame, dict containing dataframes or None
        An activity dataframe
    start_time : str or None
        The start time from
    end_time : str or None
    
    clip_activities : bool, default=False
        If set then the activities are clipped to the start and end time

    Returns
    -------
    df_d, df_a : the two subsets
    """
    # cast to pandas timestamp
    start_time = str_to_timestamp(start_time) if isinstance(start_time, str) else start_time
    end_time = str_to_timestamp(end_time) if isinstance(end_time, str) else end_time

    if df_devices is not None:
        if start_time is not None:
            dev_start_sel = (df_devices[TIME] >= start_time)
        else:
            dev_start_sel = np.full(len(df_devices), True)
        if end_time is not None:
            dev_end_sel = (df_devices[TIME] < end_time)
        else:
            dev_end_sel = np.full(len(df_devices), True)
        df_devices = df_devices[dev_start_sel & dev_end_sel]

    if df_activities is not None:
        if isinstance(df_activities, pd.DataFrame):
            df_activities = dict(df_activities=df_activities)



        for k in df_activities.keys():
            df_activity = df_activities[k]
            if start_time is not None:
                act_start_sel = ~(df_activity[END_TIME] < start_time)
            else:
                act_start_sel = np.full(len(df_activity), True)
            if end_time is not None:
                act_end_sel = ~(df_activity[START_TIME] > end_time)
            else:
                act_end_sel = np.full(len(df_activity), True)

            df_activity = df_activity[act_start_sel & act_end_sel]

            # clip activities if they extend in regions that are not in the timespan
            if not df_activity.empty and start_time is not None \
                    and (start_time < df_activity[END_TIME].iat[0] and start_time > df_activity[START_TIME].iat[0]):
                df_activity[START_TIME].iat[0] = start_time
            if not df_activity.empty and clip_activities and end_time is not None \
            and end_time < df_activity[END_TIME].iat[-1]:
                df_activity[END_TIME].iat[-1] = end_time
                
            df_activities[k] = df_activity

    if df_activities is None:
        return df_devices
    elif df_devices is None:
        return df_activities
    else:
        return df_devices, df_activities





def str_to_timestamp(val):
    """ Converts a datetime string to a panda Timestamp.
    the day-first format is used.

    Parameters
    ----------
    val : list or string

    Returns
    -------
    pd.Timestamp

    """
    return pd.to_datetime(val, dayfirst=True)

def remove_days(df_devices, df_activities, days=[], offsets=[], retain_corrections=False):
    """ Removes the given days from activities and devices and shifts the succeeding days by that amount
        forward.

    Parameters
    ----------
    df_devices : pd.DataFrame
    df_activities : pd.DataFrame
    days : list
        List of strings
    offsets : list
        Offsets that are added to the corresponding days.

    Returns
        df_devices : pd.DataFrame
        df_activities : pd.DataFrame
    """
    nr_days = len(days)
    df_devs = df_devices.copy()
    df_acts = df_activities.copy()

    # Add offsets to the days specified
    for i in range(len(days)):
        days[i] = str_to_timestamp(days[i])
        if i < len(offsets):
            days[i] = days[i] + pd.Timedelta(offsets[i])

    # sort days from last to first
    days = np.array(days)[np.flip(np.argsort(days))]
    dtypes = infer_dtypes(df_devs)


    # 1. remove iteratively the latest day and shift the succeeding part accordingly
    for day in days:
        # when day is e.g 2008-03.23 00:00:00 then day after will be 2008-03-24 00:00:00
        # these variables have to be used as only timepoints can be compared as seen below
        day_after = day + pd.Timedelta('1D')

        # Remove devices events within the selected day
        mask = (day < df_devs[TIME]) & (df_devs[TIME] < day_after)
        removed_devs = df_devs[mask].copy()
        df_devs = df_devs[~mask]

        # shift the succeeding days timeindex one day in the past
        succeeding_days = (day_after < df_devs[TIME])
        df_devs.loc[succeeding_days, TIME] = df_devs[TIME] - pd.Timedelta('1D')

        # Binary devices that change an odd amount of states in that day will have
        # a wrong state for the succeeding days until the next event
        nr_dev_triggers = removed_devs.groupby(by=[DEVICE]).count()
        for dev in nr_dev_triggers.index:
            nr = nr_dev_triggers.loc[dev, 'time']
            if nr % 2 != 0 and (dev in dtypes[BOOL] or dev in dtypes[CAT]):
                print(f'Warning: Removed odd #events: device {dev} => inconsistent device states. Correction advised!')


        # Remove activities in that day that do not extend from the previous day into the selected
        # day or extend from the selected day into the next day
        mask_act_within_day = (day < df_acts[START_TIME]) & (df_acts[END_TIME] < day_after)
        df_acts = df_acts[~mask_act_within_day]

        # Special case where one activity starts before the selected day and ends after the selected day
        mask_special = (df_acts[START_TIME] < day) & (day_after < df_acts[END_TIME])
        if mask_special.any():
            # Adjust the ending by removing one day
            df_acts.loc[mask_special, END_TIME] = df_acts[END_TIME] - pd.Timedelta('1D')

        # Shift Activities that start in or after the selected day by one day
        succeeding_days = (day <= df_acts[START_TIME]) & (day_after < df_acts[END_TIME])
        df_acts.loc[succeeding_days, START_TIME] = df_acts[START_TIME] - pd.Timedelta('1D')
        df_acts.loc[succeeding_days, END_TIME] = df_acts[END_TIME] - pd.Timedelta('1D')
        df_acts['shifted'] = False
        df_acts.loc[succeeding_days, 'shifted'] = True

        # Special case where one activity starts before the selected day and ends inside the selected day
        # and there is no activity after the selected day
        #  | day_before | Sel day      | day after
        #     |-------------|
        #    I can't just move the ending one day before as this would reverse START_TIME and END_TIME order
        # -> The last activity ending is clipped to the start of the selected day
        # TODO has to be done before any activity is shifted
        #mask_last_true = pd.Series(np.zeros(len(df_acts), dtype=np.bool_))
        #mask_last_true.iat[-1] = True
        #mask_special = (df_acts[START_TIME] < day) & (df_acts[END_TIME] <= day_after) & mask_last_true
        #if mask_special.any():
        #    assert mask_special.sum() == 1
        #    df_acts.loc[mask_special, END_TIME] = day

        df_acts = df_acts.sort_values(by=START_TIME).reset_index(drop=True)
        # Merge activities from the day_before that overlap with the shifted activities from day after
        # there are 4 cases where overlaps into the respective days have to be handled
        #                           |   db  |   sd  |   da  |=>|   db  |   sd  |   da  |=>|   db  |   sd  |   da  |
        # 1. db overlaps da            |--------|    |~|          |--------|                 |-----|~|
        #                                                              |~|
        # 2. db intersect into  da      |------|     |~~~~|         |------|                   |----|~~~~|
        #                                                               |~~~~|
        # 3. da overlaps db             |-|   |~~~~~~~|            |-|                       |-|~~~|
        #                                                       |~~~~~~~|
        # 4. da intersect into db       |---|  |~~~~~|          |---|                        |---|~~~|
        #                                                         |~~~~~|
        # 5. da intersect into db         |----||~~~~~|            |-----|                 |---|~~~|
        #                                                      |~~~~~~|
        # case 1:
        # select activities that cross the boundary between days
        mask_db_into_sd = (df_acts[START_TIME] < day) & (df_acts[END_TIME] > day) & (df_acts[END_TIME] < day_after)
        idxs = np.where(mask_db_into_sd)[0]
        assert len(idxs) <= 2
        if len(idxs) == 2:
            # case 5: case when both activities extend into each days
            # clip both to midnight
            df_acts.iat[idxs[0], 2] = day
            df_acts.iat[idxs[1], 1] = day + pd.Timedelta('1ms')
        if len(idxs) == 1:
            idx_overlapping = idxs[0]

            # Check if the overlapping activity is part of the shifted or not
            if df_acts.iat[idx_overlapping, 3]:
                # Case when the shifted activities extend into the day before the selected day
                last_unshifted_act = df_acts.loc[(df_acts[END_TIME] < day), :]\
                                .copy().sort_values(by=END_TIME, ascending=True)\
                                .iloc[-1, :]

                # clip extending activities start_time to the end_time of the first shifted activity
                df_acts.iat[idx_overlapping, 0] = last_unshifted_act[END_TIME] + pd.Timedelta('1ms')
            else:
                # Case when the previous activities extends into the selected day
                first_shifted_act = df_acts.loc[(df_acts[START_TIME] > day) & (df_acts[END_TIME] < day_after), :]\
                                            .copy().sort_values(by=START_TIME, ascending=True)\
                                            .iloc[0, :]

                # clip extending activities end_time to the start of the first shifted activity
                df_acts.iat[idx_overlapping, 1] = first_shifted_act[START_TIME] - pd.Timedelta('1ms')

        df_acts = df_acts.drop(columns='shifted')

        from pyadlml.dataset._core.devices import correct_devices
        df_devs, corrections_dev = correct_devices(df_devs, retain_corrections)

        #assert not _is_activity_overlapping(df_acts)

    if retain_corrections:
        try:
            corrections_act
        except:
            corrections_act = []
        try:
            corrections_dev
        except:
            corrections_dev = []
        return df_devs, df_acts, corrections_act, corrections_dev
    else:
        return df_devs, df_acts


def df_difference(df1: pd.DataFrame, df2: pd.DataFrame, which=None):
    """Find rows which are different between two DataFrames.

    Parameters
    ----------
    df1 : pd.DataFrame
        TODO
    df2 : pd.DataFrame
        TODO
    which : None or str, default=None
        TODO


    Returns
    -------

    """
    from pyadlml.dataset._core.devices import is_device_df
    from pyadlml.dataset._core.activities import is_activity_df

    comparison_df = df1.copy().merge(
        df2.copy(),
        indicator=True,
        how='outer'
    )
    if which is None:
        diff_df = comparison_df[comparison_df['_merge'] != 'both']
    else:
        diff_df = comparison_df[comparison_df['_merge'] == which]

    if is_activity_df(df1):
        return diff_df[[START_TIME, END_TIME, ACTIVITY]]
    elif is_device_df(df1):
        return diff_df[[TIME, DEVICE, VALUE]]
    else:
        return diff_df




def event_times(df_devices, start_time=None, end_time=None):
    """
    Parameters
    ----------
    time_array : nd.array or pd.Series or pd.DataFrame
    start_time : pd.Timestamp

    end_time : pd.Timestamp


    Returns
    -------
    res : nd.array
        Array of transformed timestamps
    start_time : pd.Timestamp

    end_time : pd.Timestamp
    """
    if isinstance(time_array, pd.DataFrame):
        time_array = time_array[TIME].values
    if isinstance(time_array, pd.Series):
        time_array = time_array.values

    # get start and end_time
    if start_time is None:
        start_time = time_array[0]
    if end_time is None:
        end_time = time_array[-1]

    # map to values between [0,1]
    res = (time_array - start_time)/(end_time - start_time)

    return res, start_time, end_time


def num_to_timestamp(val, start_time, end_time):
    """Converts value [0,1] into timestamp between start_time and end_time"""
    return start_time + val*(end_time - start_time)

def timestamp_to_num(ts, start_time, end_time):
    """Converts timestamp between start_time and end_time into value in [0,1]"""
    return float((ts - start_time)/(end_time - start_time))

def get_sorted_index(df: pd.DataFrame, rule='alphabetical', area: pd.DataFrame = None) -> np.ndarray:
    """ Returns a new dataframes index that is sorted after a specific rule

    Parameters
    ----------
    df : pd.DataFrame
        Has to contain the column 'activity' or 'device'

    rule : one of {list, ndarray, 'alphabetical', 'area', str}
        The rule on how to order the array.

        - alphabetical : the

    Returns
    -------
    new_order : ndarray
        An array with the new ordered indices for the dataframe
    """
    rule_is_iter = (isinstance(rule, list) or isinstance(rule, np.ndarray))

    # If df is a list the enumeration for that list is returned
    if (isinstance(df, list) or isinstance(df, np.ndarray)) and rule_is_iter:
        mapping = {v: k for k, v in enumerate(rule)}
        return np.vectorize(mapping.get)(df)

    df = df.copy()
    df['order'] = np.arange(len(df))
    new_order = []

    if rule_is_iter:
        mapping = {v: k for k, v in enumerate(rule)}
        order_changed = False
        try:
            df['order'] = df[ACTIVITY].map(mapping)
            order_changed = True
        except:
            pass
        try:
            df['order'] = df[DEVICE].map(mapping)
            order_changed = True
        except:
            pass
        assert order_changed, '"activity" or "device" was not present in dataframe'
    elif rule == 'alphabetical':
        if ACTIVITY in df.columns:
            df = df.sort_values(by=ACTIVITY)
        elif DEVICE in df.columns:
            df = df.sort_values(by=DEVICE)
        else:
            raise KeyError(f"Tried to sort alphabetical but no activity or device column was found, only {rule}")
    elif rule == 'value':
        # The case when the column other than ACTIVITY or order should be used
        cs = df.columns
        col = [i for i in cs if i not in [ACTIVITY, DEVICE, 'order']]
        assert len(col) == 1, 'When value is specified there must be only one remaining column.'
        if ACTIVITY in df.columns:
            df = df.sort_values(by=col[0])
        elif DEVICE in df.columns:
            df = df.sort_values(by=col[0])
        else:
            raise KeyError(f"Tried to sort alphabetical but no activity or device column was found, only {rule}")
    elif rule == 'areas':
        raise NotImplementedError
    elif isinstance(rule, str):
        assert rule in df.columns, f'The rule {rule} was not in the dataframes columns.'
        df = df.sort_values(by=rule)
    else:
        raise KeyError()
    new_order = df['order'].values

    return new_order


def extract_kwargs(func):
    """ Gets a function end sets all args to kwargs and all default values from
        the signature.
    """
    @functools.wraps(func)
    def wrapper_extract(*args, **kwargs):
        sign = inspect.signature(func).parameters

        # assign the args to the respective kwargs
        for arg_name, arg in zip(sign, args):
            kwargs[arg_name] = arg

        # set the rest of the kwargs to the default values
        for param in sign.values():
            if param.name not in kwargs.keys():
                kwargs[param.name] = param.default

        return func(**kwargs)
    return wrapper_extract


def check_order(func):
    @extract_kwargs
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        o = kwargs['order']
        assert o in ['alphabetical', 'occurence'] or isinstance(o, list) \
               or isinstance(o, np.ndarray), f'Sort rule is either "alphabetical", a custom string or a iterable. Found {o}'
        tmp = func(*args, **kwargs)
        return tmp
    return wrapper

def check_scale(func):
    @extract_kwargs
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        assert kwargs['scale'] in ['linear', 'log']
        return func(*args, **kwargs)
    return wrapper


def device_order_by(df_devs, rule='alphabetical'):
    """Return a list with ordered devices
    """
    is_iterable = isinstance(rule, list) or isinstance(rule, np.ndarray)
    if is_iterable:
        return rule

    assert rule in ['alphabetical', 'count', 'area'], f'rule{rule} not in assertion.'

    if rule == 'alphabetical':
        dev_order = df_devs[DEVICE].unique()
        dev_order.sort()
    elif rule == 'count':
        from pyadlml.dataset.stats.devices import device_order_by_count
        dev_order = device_order_by_count(df_devs)
    else:
        raise NotImplementedError('The order has still to be implemented')
    return dev_order

def activity_order_by(df_acts, rule='alphabetical'):
    """ Return a list with ordered activities
    """
    is_iterable = isinstance(rule, list) or isinstance(rule, np.ndarray)
    if is_iterable:
        return rule

    assert rule in ['alphabetical', 'duration', 'count', 'area'], f'rule{rule} not in assertion.'

    if rule == 'alphabetical':
        act_order = df_acts[ACTIVITY].unique()
        act_order.sort()
    elif rule == 'duration':
        from pyadlml.dataset.stats.activities import activity_order_by_duration
        act_order = activity_order_by_duration(df_acts)
    elif rule == 'count':
        from pyadlml.dataset.stats.activities import activity_order_by_count
        act_order = activity_order_by_count(df_acts)
    else:
        raise NotImplementedError('Area has still to be implemented.')

    return act_order

DATASET_STRINGS = [
    'casas_aruba',
    'amsterdam',
    'mitlab_1',
    'mitlab_2',
    'aras',
    'kasteren_A',
    'kasteren_B',
    'kasteren_C',
    'tuebingen_2019',
    'uci_ordonezA',
    'uci_ordonezB',
    'act_assist',
    'dump'
]
def fetch_data_by_string(dataset: str, cache=True, identifier=None):
    """ Fetches data based on a given string representation

    Parameters
    ----------
    dataset: str

    identifier: str, default=None
        Identifies a custom dataset in the data home. In the case
        of an act_assist dataset this is the folder name. In the case
        of a dumped dataframe with pyadlml.io.dump this is the name.

    Returns
    -------
    data : pyadlml.dataset.obj
    """

    from pyadlml.dataset import fetch_amsterdam, \
                                fetch_mitlab, fetch_aras, fetch_kasteren_2010, \
                                fetch_casas_aruba, fetch_tuebingen_2019, \
                                fetch_uci_adl_binary, load_act_assist
    if dataset == DATASET_STRINGS[0]:
        data = fetch_casas_aruba(cache=cache)
    elif dataset == DATASET_STRINGS[1]:
        data = fetch_amsterdam(cache=cache)
    elif dataset == DATASET_STRINGS[2]:
        data = fetch_mitlab(subject='subject1', cache=cache)
    elif dataset == DATASET_STRINGS[3]:
        data = fetch_mitlab(subject='subject2', cache=cache)
    elif dataset == DATASET_STRINGS[4]:
        data = fetch_aras(cache=cache)
    elif dataset == DATASET_STRINGS[5]:
        data = fetch_kasteren_2010(house='A', cache=cache)
    elif dataset == DATASET_STRINGS[6]:
        data = fetch_kasteren_2010(house='B', cache=cache)
    elif dataset == DATASET_STRINGS[7]:
        data = fetch_kasteren_2010(house='C', cache=cache)
    elif dataset == DATASET_STRINGS[8]:
        data = fetch_tuebingen_2019(cache=cache)
    elif dataset == DATASET_STRINGS[9]:
        data = fetch_uci_adl_binary(subject='OrdonezA', cache=cache)
    elif dataset == DATASET_STRINGS[10]:
        data = fetch_uci_adl_binary(subject='OrdonezB', cache=cache)
    elif dataset == DATASET_STRINGS[11]:
        data = load_act_assist(identifier)
    elif dataset == DATASET_STRINGS[12]:
        from pyadlml.dataset._core.devices import is_device_df
        from pyadlml.dataset.io import load
        from pyadlml.dataset._core.activities import is_activity_df
        df_lst = load(identifier)

        df_devs = None
        df_acts = None
        for df in df_lst:
            if is_device_df(df):
                df_devs = df
            elif is_activity_df(df):
                df_acts = df
            else:
                raise ValueError('At least one of the dataframes has to be a device dataframe')
        assert df_devs is not None and df_acts is not None
        data = Data(df_acts, df_devs)

    else:
        raise ValueError(f'No suitable option specified: {dataset}.\nAvailable are {DATASET_STRINGS}')

    return data


def get_dev_row_where(df, time, dev, state):
    time = pd.Timestamp(time)
    df = df.copy().reset_index().set_index(TIME)
    mask = (df[DEVICE] == dev) \
           & (df.index == time) \
           & (df[VALUE] == state)
    df = df.reset_index().set_index('index')
    return df[mask.values].copy()

def get_dev_rows_where(df_devs, rows):
    res = [get_dev_row_where(df_devs, r[0], r[1], r[2]) for r in rows]
    return pd.concat(res)

def append_devices(df_devs, rows):
    df = pd.DataFrame(data=rows, columns=[TIME, DEVICE, VALUE])
    df[TIME] = pd.to_datetime(df[TIME])
    return pd.concat([df_devs, df])

def remove_devices(df_devs, rows):
    idx_to_drop = [get_dev_row_where(df_devs, r[0], r[1], r[2]).index[0] for r in rows]
    return df_devs.drop(index=idx_to_drop)



"""
    this is an abstract class of an implemented algorithm like hmm
    model is used for benchmarks
    model is also called from hmm events
"""
import datetime

from hassbrain_algorithm.controller import Controller
import joblib
import copy

MD_FILE_NAME = "model_%s.joblib"


class Model(object):

    def __init__(self, name, controller):
        self._bench = None
        self._ctrl = controller # type: Controller
        self._model_name = MD_FILE_NAME

        """
        these are callbacks that are can be accessed from third party classes
        for example the Benchmark to execute sth. during each training step.
        The model such as a hmm should call these methods after each step
        """
        self._callbacks = []

        """
        these hashmaps are used to decode and encode in O(1) the numeric based
        values of states and observations to the string based representations 
        """
        self._obs_lbl_hashmap = {}
        self._obs_lbl_rev_hashmap = {}
        self._state_lbl_hashmap = {}
        self._state_lbl_rev_hashmap = {}

    def encode_state_lbl(self, label):
        return self._state_lbl_hashmap[label]

    def decode_state_lbl(self, ide):
        return self._state_lbl_rev_hashmap[ide]

    def encode_obs_lbl(self, label, state):
        """
        returns the id of a sensor given a label
        :param label:
        :return:
        """
        return self._obs_lbl_hashmap[label][state]

    def decode_obs_label(self, ide):
        """
        retrieves the label given a sensor id
        :param id:
        :return:
        """
        return self._obs_lbl_rev_hashmap[ide]

    def get_state_label_list(self):
        lst = []
        for key, val in self._state_lbl_rev_hashmap.items():
            lst.append(key)
        return lst

    def get_obs_label_list(self):
        lst = []
        for key, val in self._obs_lbl_hashmap.items():
            lst.append(key)
        return lst


    def obs_lbl_seq2enc_obs_seq(self, obs_seq):
        """
        generates from labels and observations
        :param obs_seq:
            list of tupels with a label and a state of the observation
            example: [('binsens.motion', 0), ('binsens.dis', 1), ... ]
        :return:
            list of numbers of the encoded observations
        """
        enc_obs_seq = []
        for tupel in obs_seq:
            enc_obs_seq.append(self._obs_lbl_hashmap[tupel[0]][tupel[1]])
        return enc_obs_seq

    def append_method_to_callbacks(self, callback_method):
        self._callbacks.append(callback_method)

    def set_train_loss_callback(self):
        self._callbacks.append(self._train_loss_callback)

    def _train_loss_callback(self, *args):
        """
        hass to format the callback from the real model into an appropriate output for
        the benchmark method bench.train_loss_callback()
        :return:
        """
        raise NotImplementedError

    def register_benchmark(self, bench):
        self._bench = bench

    def generate_file_path(self, path_to_folder, filename):
        key = 1
        #name = path_to_folder + "/" + filename + "_%s.joblib"%(key)
        name = path_to_folder + "/" + filename
        return name

    def save_model(self, path_to_folder, filename):
        full_file_path = self.generate_file_path(path_to_folder, filename)
        import os
        if not os.path.exists(path_to_folder):
            os.makedirs(path_to_folder)

        joblib.dump(self, full_file_path)

    def load_model(self, path_to_folder, filename):
        name = self.generate_file_path(path_to_folder, filename)
        return joblib.load(name)


    def set_name(self, name):
        self._name = name

    def get_name(self):
        return self._name

    def model_init(self, dataset, state_list=None):
        """
        initialize model on dataset
        :param dataset:
        :param state_list:
        :param location_data:
            a list conatining information how the smart home is set up
            loc_data = [ { "name" : "loc1", "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_floor', 'binary_sensor.motion_mirror'],
            }, ... }
        :return:
        """

        self.gen_hashmaps(dataset)

        if state_list is not None:
            self._state_list = dataset.encode
            # todo where is this used and what is state list set to
            raise ValueError
        else:
            self._state_list = dataset.get_state_list()



        self._observation_list = dataset.get_obs_list()

        self._model_init(dataset)

    def gen_hashmaps(self, dataset):
        self._obs_lbl_hashmap = dataset.get_obs_lbl_hashmap()
        self._obs_lbl_rev_hashmap = dataset.get_obs_lbl_reverse_hashmap()
        self._state_lbl_hashmap = dataset.get_state_lbl_hashmap()
        self._state_lbl_rev_hashmap = dataset.get_state_lbl_reverse_hashmap()

    def are_hashmaps_created(self):
        return self._obs_lbl_hashmap is None \
               or self._obs_lbl_rev_hashmap is None \
               or self._state_lbl_hashmap is None \
               or self._state_lbl_rev_hashmap is None


    def register_act_info(self, act_data):
        self._act_data = self._encode_act_data(act_data)

    def register_loc_info(self, loc_data):
       self._loc_data = self._encode_location_data(loc_data)

    def _encode_act_data(self, act_data):
        """
        encodes the device names and activity names into numbers that the models
        can understand
        :param loc_data:
            is a list of loc data
            example:
            [ { "name" : "loc1", "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_floor', 'binary_sensor.motion_mirror'],
            },  ... ]
        :return:
            a list of location data, with the same structure correct encoded labels
             [ { "name" : "loc1",
                "activities" : [ 1 ],
                "devices" : [3, 7],
            },  ... ]
        """
        for activity in act_data:
            activity["name"] = self.encode_state_lbl(activity["name"])
        return act_data

    def sort_act_data(self):
        """
        sorts the synthetic generated activity data and returns a sorted list
        :return:
        """
        # list of lists, with each sublist representing a day
        # daybins[0] contains all activities for sunday, ...
        day_bins = [[],[],[],[],[],[],[]]
        for syn_act in self._act_data:
            day_bins[syn_act["day_of_week"]].append(syn_act)
        res_list = []
        for day in day_bins:
            for i in range(len(day)):
                min_idx = self._sort_get_min_idx(day)
                res_list.append(day[i])
        return res_list

    def _sort_get_min_idx(self, lst):
        """
        returns the index of the smallest element
        :param lst:
        :return:
        """
        min_idx = 0
        for i in range(1, len(lst)):
            if lst[i]['start'] < lst[min_idx]['start']:
                min_idx = i
        return min_idx






    def _encode_location_data(self, loc_data):
        """
        encodes the device names and activity names into numbers that the models
        can understand
        :param loc_data:
            is a list of loc data
            example:
            [ { "name" : "loc1", "activities" : ['cooking'],
            "devices" : ['binary_sensor.motion_floor', 'binary_sensor.motion_mirror'],
            },  ... ]
        :return:
            a list of location data, with the same structure correct encoded labels
             [ { "name" : "loc1",
                "activities" : [ 1 ],
                "devices" : [3, 7],
            },  ... ]
        """
        for location in loc_data:
            new_act_list = []
            for activity in location['activities']:
                # todo the use of str vs. int is a hint that there is an
                # inconsistent use in the encoding and decoding in state labels
                new_act_list.append(str(self.encode_state_lbl(activity)))
            location['activities'] = new_act_list

            new_dev_list = []
            for device in location['devices']:
                new_dev_list.append(self.encode_obs_lbl(device, 0))
                new_dev_list.append(self.encode_obs_lbl(device, 1))
            location['devices'] = new_dev_list
        return loc_data




    def _model_init(self, dataset, location_data):
        """
        this method has to be overriden by child classes
        :return:
        """
        pass

    def train(self, dataset):
        """
         gets a dataset and trains the model on the data
         Important !!!!
         during the training the hashmaps of the model have to be filled up
        :param dataset:
        :return:
        """
        pass


    def get_state(self, seq):
        """
        returns the state the model is in given an observation sequence
        :return:
        """
        pass

    def save_visualization(self, path_to_file):
        """
        save a visualization of the model to the given filepath
        :param path_to_file:
        :return:
        """
        pass

    def draw(self):
        """
         somehow visualize the model
        :return: an image png or jpg
        """
        pass


    def classify_multi(self, obs_seq):
        """
        gets an observation sequence (at most times this is a window) and returns
        the most likely states
        :param obs_seq:
        :return: np array with the most likely state
        """
        # encode obs_seq
        obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        # array full of tuples with state_label and corresp. score
        pred_state_arr = self._classify_multi(obs_seq)

        # decode state seq
       # print('~'*10)
        act_score_dict = {}
        for i in range(0,len(pred_state_arr)):
            label = pred_state_arr[i][0]
            score = pred_state_arr[i][1]
            act_score_dict[self._state_lbl_rev_hashmap[label]] = score
        return act_score_dict

    def classify(self, obs_seq):
        """
        gets an observation sequence (at most times this is a window) and returns
        the most likely states
        :param obs_seq:
        :return: single state with most likely state
        """
        # encode obs_seq
        obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        pred_state = self._classify(obs_seq)

        # decode state seq
        res = self._state_lbl_rev_hashmap[pred_state]
        return res

    def predict_next_obs(self, obs_seq):
        """

        :param args:
        :return:
            the next observation
        """
        obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        pred_obs = self._predict_next_obs(obs_seq)
        # hack todo make a design change
        label = self._obs_lbl_rev_hashmap[pred_obs]
        if pred_obs % 2 == 0:
            return (label, 0)
        else:
            return (label, 1)

    def predict_prob_xnp1(self, obs_seq):
        """
        computes the probabilities of all observations to be the next
        :param obs_seq:
        :return:
            dictionary like: "{ 'sensor_name' : { 0 : 0.123, 1: 0.789 } , ... }"
        """
        obs_seq = self.obs_lbl_seq2enc_obs_seq(obs_seq)
        arr = self._predict_prob_xnp1(obs_seq)
        res_dict = copy.deepcopy(self._obs_lbl_hashmap)
        #print('*'*10)
        #print(arr)
        #print(res_dict.__eq__(self._obs_lbl_hashmap))
        #print(self._obs_lbl_rev_hashmap)
        #print('*'*10)
        for i, prob in enumerate(arr):
            label = self._obs_lbl_rev_hashmap[i]
            # if the index is even, than the observation is 'turn on'
            if i % 2 == 0:
                res_dict[label][0] = prob
            else:
                res_dict[label][1] = prob
        return res_dict

    def _classify(self, obs_seq):
        pass

    def _classify_multi(self, obs_seq):
        pass

    def _predict_next_obs(self, obs_seq):
        """
        has to return an array containing all the probabilities of the observations
        to be next
        :param obs_seq:
        :return:
        """
        pass

    def _predict_prob_xnp1(self, obs_seq):
        pass



import math
from collections import OrderedDict
from .util import _style_colorbar, _dyn_y_label_size
import plotly
import pandas as pd
import numpy as np

from plotly.subplots import make_subplots
import plotly.graph_objects as go
from plotly.express import IdentityMap
from plotly.express._core import apply_default_cascade, infer_config, \
    get_label, make_trace_kwargs, get_decorated_label, init_figure, make_trendline_spec, configure_axes, \
    configure_animation_controls, process_dataframe_timeline, one_group, \
    get_groups_and_orders

#from plotly.subplots import _subplot_type_for_trace_type, _set_trace_grid_reference
from plotly.validators.choropleth import ColorscaleValidator
from pyadlml.constants import DEVICE, TIME, ACTIVITY, START_TIME, END_TIME, VALUE, NUM, BOOL, \
    CAT, STRFTIME_DATE
import plotly.express as px

from pyadlml.dataset._core.devices import device_events_to_states
from pyadlml.dataset.matplotlib.util import format_device_labels
from pyadlml.dataset.plotly.activities import _set_compact_title
from pyadlml.dataset.stats.acts_and_devs import contingency_table_states, contingency_table_events
from pyadlml.dataset.util import select_timespan, df_difference, activity_order_by, device_order_by, infer_dtypes


__all__ = ['activities_and_devices', 'contingency_states', 'contingency_events']


def _plot_device_states_into_fig(fig: go.Figure, df_devs: pd.DataFrame,  df_devs_usel: pd.DataFrame,
                                 df_devs_outside: pd.DataFrame, dev_order: list, st=None, et=None) -> go.Figure:
    """
    Parameters
    ----------
    df_devs_outside : list of dicts
        Each dictionary contains the key 'df' and a 'color' and a 'opacity'


    """
    EVENT_COLOR = 'Black'
    #marker = dict(size=5, symbol=42, line=dict(color=EVENT_COLOR, width=1))

    df_devs = df_devs.copy()\
                        .sort_values(by=TIME)\
                        .reset_index(drop=True)
    devs = df_devs[DEVICE].unique()
    dtypes = infer_dtypes(df_devs)
    df_devs = device_events_to_states(df_devs, extrapolate_states=True,
                                      st=st, et=et).reset_index(drop=True)

    if df_devs_usel is not None:
        mark_selected = {}
        df_devs_usel = df_devs_usel.rename(columns={TIME: START_TIME})
        if END_TIME not in df_devs_usel.columns:
            tmp = df_devs.copy()
            comp_df = tmp[[START_TIME, DEVICE, VALUE]]\
                      .merge(df_devs_usel, indicator=True, how='left')
            mask = (comp_df['_merge'] == 'both')
            tmp2 = tmp[mask]
            df_devs_usel = tmp2
        df_devs_usel = _endtime_to_offset(df_devs_usel, replace=False)
        df_devs_usel[VALUE] = df_devs_usel[VALUE].map({True: 'on', False: 'off'})


    df_devs = _endtime_to_offset(df_devs, replace=False)

    # A mapping from device to data index
    data_dict = {}
    j = 0
    first_bool = True
    for i, dev in enumerate(devs):
        df = df_devs[df_devs[DEVICE] == dev].copy()

        if dev in dtypes[BOOL]:
            df[VALUE] = df[VALUE].map({True: 'on', False: 'off'})
            df_on = df[df[VALUE] == 'on']
            df_off = df[df[VALUE] == 'off']

            if df_devs_usel is not None:
                mark_selected[dev] = {}
                comp_df = df_on.copy().merge(df_devs_usel, indicator=True, how='left')
                tmp = np.where((comp_df['_merge'] == 'both').values)[0]
                comp_df2 = df_off.copy().merge(df_devs_usel, indicator=True, how='left')
                tmp2 = np.where((comp_df2['_merge'] == 'both').values)[0]
                mark_selected[dev]['on'] = tmp
                mark_selected[dev]['off'] = tmp2

            COL_TRUE = 'teal'
            COL_FALSE = 'lightgray'
            hover_template = '%{y}<br>' + \
                             'Start_time: %{base|' + STRFTIME_DATE + '}<br>' + \
                             'End_time: %{customdata[1]|' + STRFTIME_DATE + '}<br>' + \
                             'Duration: %{customdata[0]}<br>' + \
                             'State: %{customdata[2]}<extra></extra>'

            def create_trace(df, state):
                df['diff'] = (df[END_TIME] - df[START_TIME]).astype(str)
                cd = df[['diff', END_TIME]].values
                # Add state information to custom_data
                vals = np.expand_dims(np.full(cd.shape[0], (state == 'on')), axis=1)
                cd = np.hstack([cd, vals])
                marker_color = COL_TRUE if state == 'on' else COL_FALSE
                return go.Bar(name=state,
                              base=df[START_TIME],
                              meta=dev,
                              x=df['offset'],
                              y=df[DEVICE],
                              marker_color=marker_color,
                              customdata=cd,
                              legendgroup=state,
                              orientation='h',
                              width=0.3,
                              textposition='auto',
                              showlegend=first_bool,
                              hovertemplate=hover_template,
                )

            trace_off = create_trace(df_off, 'off')
            trace_on = create_trace(df_on, 'on')
            data_dict[dev] = [len(fig.data), len(fig.data)+1]
            fig.add_traces([trace_on, trace_off])

            first_bool = False
        elif dev in dtypes[CAT]:
            categories = df.loc[df[DEVICE] == dev, VALUE].unique()
            for cat in categories:
                #values = df.loc[(df[VAL] == cat), [col_bar_start, col_bar_len]].values.tolist()
                #ax.broken_barh(values, (i-0.25, 0.5), facecolors=tab(j), label=dev + ' - ' + cat)
                j += 1
            raise NotImplementedError

        elif dev in dtypes[NUM]:
            values = pd.to_numeric(df[VALUE])
            values = (values-values.min())/(values.max() - values.min())*0.5
            values = values + i - 0.25
            #ax.plot(df['num_st'], values, color=color_num, linestyle='--', marker='o')
            raise NotImplementedError

        # Create user_selection for each trace
        if df_devs_usel is not None:
            if dev in dtypes[BOOL]:
                fig.update_traces(selector=dict(meta=dev, name='off'),
                           selectedpoints=mark_selected[dev]['off'],
                           selected={'marker': {'opacity': 1.0, 'color': COL_FALSE}},
                           unselected={'marker': {'opacity': 0.3, 'color': COL_FALSE}})
                fig.update_traces(selector=dict(meta=dev, name='on'),
                           selectedpoints=mark_selected[dev]['on'],
                           selected={'marker': {'opacity': 1.0, 'color': COL_TRUE}},
                           unselected={'marker': {'opacity': 0.3, 'color': COL_TRUE}})
            else:
                raise NotImplementedError


    fig.update_layout(yaxis_type='category')
    fig.update_yaxes(categoryorder='array', categoryarray=np.flip(dev_order))

    return fig


def _plot_device_events_into_fig(fig: go.Figure, df_devs: pd.DataFrame,  df_devs_usel: pd.DataFrame,
                                 df_devs_outside: pd.DataFrame, dev_order: list,
                                 marker_height=5,
                                 ) -> go.Figure:
    """
    Parameters
    ----------
    df_devs_outside : list of dicts
        Each dictionary contains the key 'df' and a 'color' and a 'opacity'


    """
    # Enable webgl rendering
    scatter = go.Scattergl if len(df_devs) > 15000 else go.Scatter
    EVENT_COLOR = 'Black'
    hover_template = '%{y}<br>%{x|' + STRFTIME_DATE + '}<br>Event: %{customdata} <extra></extra>'


    marker = dict(size=marker_height, symbol=42, line=dict(color=EVENT_COLOR, width=1))
    fig.update_layout(yaxis_type='category')
    fig.update_yaxes(categoryorder='array', categoryarray=np.flip(dev_order))
    fig.add_trace(scatter(
        mode='markers', y=df_devs[DEVICE], x=df_devs[TIME],
        customdata=df_devs[VALUE],
        hovertemplate=hover_template,
        showlegend=False, marker=marker))

    if not df_devs_outside.empty:
        marker['opacity'] = 0.1
        marker['line']['color'] = 'Grey'
        fig.add_trace(scatter(mode='markers', y=df_devs_outside[DEVICE],
                                 x=df_devs_outside[TIME], showlegend=False,
                                 marker=marker, hoverinfo='skip')
                      )

    # Create user_selection for each trace
    if df_devs_usel is not None:
        df_devs_usel = df_devs_usel[[TIME, DEVICE, VALUE]]
        # Get the indices in the trace where
        comp_df = df_devs.copy().merge(df_devs_usel, indicator=True, how='left')
        mark_selected = np.where((comp_df['_merge'] == 'both').values)[0]
        unselected = {'marker': {'opacity': 0.2, 'color': EVENT_COLOR}}
        selected = {'marker': {'opacity': 1.0, 'color': 'Red'}}
        fig.update_traces(selector=dict(type="scatter"),
                          selectedpoints=mark_selected,
                          selected=selected, unselected=unselected)

    return fig


def _determine_start_and_end(df_acts: dict, df_devs: pd.DataFrame, st: pd.Timestamp, et: pd.Timestamp):
    """ Determine the start and endpoint with regard to optional given parameters
    """
    if isinstance(df_acts, pd.DataFrame):
        act_st = df_acts[START_TIME].iloc[0]
        act_et = df_acts[END_TIME].iloc[-1]
    elif isinstance(df_acts, dict):
        act_st = df_acts.get_min_start_time()
        act_et = df_acts.get_max_end_time()

    data_st = min(df_devs[TIME].iloc[0], act_st) - pd.Timedelta('1ms')
    data_et = max(df_devs[TIME].iloc[-1], act_et) + pd.Timedelta('1ms')
    if st is None:
        draw_start_line = False
        st = data_st
    else:
        # Draw the line only if the given start is the right-most
        draw_start_line = (max(st, data_st) == st)
        st = max(st, data_st)
    if et is None:
        draw_end_line = False
        et = data_et
    else:
        # Draw the line only if the given end is the left-most
        draw_end_line = (min(et, data_et) == et)
        et = min(et, data_et)
    return st, et, draw_start_line, draw_end_line



def ActDictLoop(func):
    def func_wrapper(*args, **kwargs):
        if isinstance(df_acts_sel, pd.DataFrame):
            df_acts_sel = ActivityDict(dict(tmp=df_acts_sel))
        if isinstance(df_acts, pd.DataFrame):
            df_acts = ActivityDict(dict(tmp=df_acts))
            
    return 

@ActDictLoop("df_acts_sle, df_acts")
def act_difference(df_acts_sel, df_acts, st, et):
    """ Get outside activities with enveloping activities correctly clipped

    Parameters
    ----------
    df_acts_sel : pd.DataFrame or ActivityDict or List
        asdf
    df_acts : pd.DataFrame or ActivityDict or List
    st : str
    et : str

    Returns
    -------

    """
    df_acts_outside = df_difference(df_acts_sel, df_acts)\
                      .sort_values(by=START_TIME)\
                      .reset_index(drop=True)

    # ao contains both the cut up and the non-cut activities, adjust the start activity
    if not df_acts_outside.empty:
        df_acts_outside[ACTIVITY] = 'not selected'

        # Both the old time and the new start_time are in df_acts_outside
        # Get entry of outside corresponding to start_time
        act_st_split_et = df_acts_outside[df_acts_outside[START_TIME] == st]
        if not act_st_split_et.empty:
            idxs_acts_st_split = df_acts_outside[(df_acts_outside[END_TIME] == act_st_split_et.iat[0, 1])].index
            df_acts_outside.iat[idxs_acts_st_split[0], 1] = st - pd.Timedelta('1ms')
            df_acts_outside = df_acts_outside.drop(index=idxs_acts_st_split[1])

            df_acts_outside = df_acts_outside.reset_index(drop=True)


        act_st_split_st = df_acts_outside[df_acts_outside[END_TIME] == et]
        if not act_st_split_st.empty:
            idxs_acts_et_split = df_acts_outside[(df_acts_outside[START_TIME] == act_st_split_st.iat[0, 0])].index
            df_acts_outside.iat[idxs_acts_et_split[1], 0] = et - pd.Timedelta('1ms')
            df_acts_outside = df_acts_outside.drop(index=idxs_acts_et_split[0])

            df_acts_outside = df_acts_outside.reset_index(drop=True)



    return df_acts_outside


def _plot_selected_activity_marker(fig, df):
    """ Plot markers at the midpoint of selected activities
    """
    df = df[[START_TIME, END_TIME, ACTIVITY]].copy()
    cd = df.values
    diff = df[END_TIME] - df[START_TIME]
    df['mid_point'] = df[START_TIME] + diff/2
    y_label = 'Selected Activity mark'
    df['y'] = y_label

    marker = dict(size=5, symbol=5, line=dict(color='Red', width=1))
    hover_template = 'Activity: %{customdata[2]}<br>' + \
                     'Start_time: %{customdata[0]|' + STRFTIME_DATE + '}<br>' + \
                     'End_time: %{customdata[1]|' + STRFTIME_DATE + '}<br>' + \
                     '<extra></extra>'

    fig.add_trace(go.Scatter(
        mode='markers', y=df['y'], x=df['mid_point'],
        customdata=cd,
        marker=marker,
        hovertemplate=hover_template,
        showlegend=False))
    y_axis_order = [y_label] + fig.layout.yaxis.categoryarray.tolist()
    fig.update_yaxes(categoryarray=y_axis_order)

    return fig


def _plot_selected_device_marker(fig, df, df_devs, states=False):
    df = df[[TIME, DEVICE, VALUE]].copy()
    df_devs = df_devs[[TIME, DEVICE, VALUE]].copy()\
        .sort_values(by=TIME)\
        .reset_index(drop=True)

    y_label = 'Selected Device mark'
    df['y'] = y_label

    if states:
        # Set the categorical and boolean markers to midpoint of state
        dtypes = infer_dtypes(df)
        df[END_TIME] = df[TIME].copy()
        for dev in dtypes[BOOL] + dtypes[CAT]:
            df_tmp = df_devs[df_devs[DEVICE] == dev].copy()
            df_tmp[END_TIME] = df_tmp[TIME].shift(-1)
            mask = (dev == df[DEVICE])
            times = df.loc[mask, TIME]
            vals = df_tmp[df_tmp[TIME].isin(times)]
            df.at[mask, END_TIME] = vals

        df[START_TIME] = df[TIME]
        df[TIME] = df[TIME] + (df[END_TIME] - df[TIME])/2
        custom_data = df[[DEVICE, START_TIME, END_TIME]].values
        hover_template = 'Device: %{customdata[0]}<br>' + \
                         'Start_time: %{customdata[1]|' + STRFTIME_DATE + '}<br>' + \
                         'End_time: %{customdata[2]|' + STRFTIME_DATE + '}<br>' + \
                         '<extra></extra>'
    else:
        custom_data = df[DEVICE]
        hover_template = 'Device: %{customdata}<br>Time: %{x}<extra></extra>'
    marker = dict(size=5, symbol=5, line=dict(color='Red', width=1))
    fig.add_trace(go.Scatter(
        mode='markers', y=df['y'], x=df[TIME],
        customdata=custom_data,
        marker=marker,
        hovertemplate=hover_template,
        showlegend=False))
    y_axis_order = [y_label, *list(fig.layout.yaxis.categoryarray)]
    fig.update_yaxes(categoryarray=y_axis_order)
    return fig


def activities_and_devices(df_acts, df_devs, states=False, st=None, et=None,
                           act_order='alphabetical', dev_order='alphabetical',
                           df_acts_usel=None, df_devs_usel=None, devs_usel_state=None,
                           height=350
                           ):

    # Determines if device markers indicate the event or a states midpoint
    if states:
        devs_usel_state = states if devs_usel_state is None else devs_usel_state
    else:
        devs_usel_state = False
    assert isinstance(devs_usel_state, bool), 'devs_usel_state has to be set with either a boolean or None'

    # Copy data
    if isinstance(df_acts, pd.DataFrame): 
        df_acts = df_acts.copy().reset_index(drop=True)  # TODO
        df_acts = ActivityDict({'subject':df_acts})
    
    df_devs = df_devs.copy()

    # Determine the start and endpoint with regard to optional given parameters
    st, et, draw_start_line, draw_end_line = _determine_start_and_end(df_acts, df_devs, st, et)

    # Select the active displayed parts
    df_devs_sel, df_acts_sel = select_timespan(df_devs, df_acts, st, et, clip_activities=True)

    # Get the y-axis and label order
    act_order = activity_order_by(df_acts_sel.concat(), rule=act_order)
    dev_order = device_order_by(df_devs_sel, rule=dev_order)


    # determinte visual properties
    nr_devs = len(dev_order)
    marker_height = 3 if nr_devs > 15 else 5
    y_label_size = _dyn_y_label_size(height, nr_devs)

    # Reconstruct outside parts
    df_devs_outside = df_difference(df_devs_sel, df_devs)
    df_acts_outside = act_difference(df_acts_sel, df_acts, st, et)

    title = 'Activities and device events'
    fig = go.Figure()
    if states:
        fig = _plot_device_states_into_fig(fig, df_devs_sel, df_devs_usel,
                                           df_devs_outside, dev_order, st, et)
    else:
        fig = _plot_device_events_into_fig(fig, df_devs_sel, df_devs_usel, df_devs_outside, dev_order, marker_height)

    fig = _plot_activities_into_fig(fig, df_acts_sel, df_acts_usel, df_acts_outside, act_order)

    # Plot markers into trace
    if df_acts_usel is not None:
        fig = _plot_selected_activity_marker(fig, df_acts_usel)

    if df_devs_usel is not None:
        fig = _plot_selected_device_marker(fig, df_devs_usel, df_devs, devs_usel_state)



    if draw_start_line:
        fig.add_vline(x=st, line_width=1, line_color="Grey")
    if draw_end_line:
        fig.add_vline(x=et, line_width=1, line_color="Grey")

    _set_compact_title(fig, title)
    fig.update_yaxes(title=None, fixedrange=True, tickfont=dict(size=y_label_size),
                     ticklabeloverflow='allow'
                     )
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)

    return fig


def _endtime_to_offset(df_act, replace=True):
    """Compute the end_time as numerical offset in ms"""
    x_start = pd.to_datetime(df_act[START_TIME])
    x_end = pd.to_datetime(df_act[END_TIME])
    col_label = END_TIME if replace else 'offset'
    df_act[col_label] = (x_end - x_start).astype("timedelta64[ms]")
    return df_act


def _plot_activities_into_fig(fig, df_acts: pd.DataFrame, df_acts_usel: pd.DataFrame,
                              df_acts_outside: pd.DataFrame, act_order: list, y_label='Activity',
                              color_discrete_map=None) -> go.Figure:
    """ Manually plot timeline from plotly.express.make_figure
        fig = px.timeline(df_acts, x_start=START_TIME, x_end=END_TIME, y='y_label',
                              color=ACTIVITY)

    Parameters
    ----------
    fig : obj.Figure

    df_acts : pd.DataFrame

    mask_unselected : pd.Series
        A mask that indicates the datapoints whichs opacity should be lowered.
    """

    df_acts['y_label'] = y_label
    df_acts = df_acts[['y_label', START_TIME, END_TIME, ACTIVITY]]
    df_acts['dur'] = (df_acts[END_TIME] - df_acts[START_TIME]).astype(str)

    df_acts_outside['y_label'] = y_label
    df_acts_outside = df_acts_outside[['y_label', START_TIME, END_TIME, ACTIVITY]]
    df_acts_outside['dur'] = (df_acts_outside[END_TIME] - df_acts_outside[START_TIME]).astype(str)

    if df_acts_usel is not None:
        df_acts_usel['y_label'] = y_label
        df_acts_usel = df_acts_usel[['y_label', START_TIME, END_TIME, ACTIVITY]]
        df_acts_usel['dur'] = (df_acts_usel[END_TIME] - df_acts_usel[START_TIME]).astype(str)
        df_acts_usel = _endtime_to_offset(df_acts_usel)

    args = dict(data_frame=df_acts,
                x_start=START_TIME,
                x_end=END_TIME,
                y='y_label',
                color=ACTIVITY,
                color_discrete_map=color_discrete_map,
                category_orders={'activity': []},
                #categoryarray=category_orders,

                # Necessary to work
                template=None,
                color_discrete_sequence=None,
                hover_data=None,
    )

    trace_patch = dict(textposition="auto", orientation="h")
    layout_patch = dict(barmode="overlay")

    # Process args
    apply_default_cascade(args)
    args = process_dataframe_timeline(args)

    trace_specs, grouped_mappings, sizeref, _ = infer_config(
        args, go.Bar, trace_patch, layout_patch
    )

    grouper = [x.grouper or one_group for x in grouped_mappings] or [one_group]
    grouped = args["data_frame"].groupby(grouper, sort=False)

    # Grouped is already ordered TODO debug
    _, orders = get_groups_and_orders(args, grouper)
    #orders, sorted_group_names = get_orderings(args, grouper, grouped)

    # Assign colors to the groups
    for val in act_order:
        m = grouped_mappings[0]
        if val not in m.val_map:
            m.val_map[val] = m.sequence[len(m.val_map) % len(m.sequence)]


    trace_lst = []
    for group_name in act_order:
        df_sel_act = grouped.get_group(group_name)
        act_name = group_name

        # Create the trace
        trace = go.Bar(name=act_name)
        trace.update(legendgroup=act_name, showlegend=True,
                     alignmentgroup=True, offsetgroup=act_name)

        # Init subplot row/col
        trace._subplot_row = 1
        trace._subplot_col = 1

        m = grouped_mappings[0]
        trace_color = m.val_map[group_name]
        m.updater(trace, trace_color)
        hover_template = '<b>' + act_name + '</b><br>'\
                         + 'Start_time: %{base|' + STRFTIME_DATE + '}<br>' \
                         + 'End_time: %{x| ' + STRFTIME_DATE + '}' \
                         + '<br>Duration: %{customdata}<extra></extra>'
        trace.update(dict(textposition='auto', orientation='h',
                          base=df_sel_act[START_TIME],
                          x=df_sel_act[END_TIME],
                          y=df_sel_act['y_label'],
                          customdata=df_sel_act['dur'],
                          hovertemplate=hover_template
        ))

        trace_lst.append(trace)

        # Create user_selection for each trace
        if df_acts_usel is not None:
            # Get the indices in the trace where
            comp_df = df_sel_act.copy().merge(df_acts_usel, indicator=True, how='left')
            mark_selected = np.where((comp_df['_merge'] == 'both').values)[0]
            unselected = {'marker': {'opacity': 0.3, 'color': trace_color}}
            selected = {'marker': {'opacity': 1.0, 'color': trace_color}}
            trace.update(selectedpoints=mark_selected, selected=selected, unselected=unselected)

    if not df_acts_outside.empty:
        # Convert to bar
        act_name = 'not selected'
        df_sel_act = _endtime_to_offset(df_acts_outside)

        # Create the trace
        trace = go.Bar(name=act_name,
                       base=df_sel_act[START_TIME],
                       hoverinfo='skip',
                       marker=dict(color='rgba(58, 71, 80, 0.6)', opacity=0.2),
                       x=df_sel_act[END_TIME],
                       y=df_sel_act['y_label'],
                       customdata=df_sel_act['dur'],
                       textposition='auto',
                       orientation='h',
                       legendgroup=act_name,
                       showlegend=False,
                       alignmentgroup=True,
                       offsetgroup=act_name,
        )
        trace._subplot_row = 1
        trace._subplot_col = 1
        trace_lst.append(trace)

    fig = make_subplots(
        rows=1,
        cols=1,
        shared_xaxes='all',
        shared_yaxes='all',
        start_cell="bottom-left",
        horizontal_spacing=0.02,
        vertical_spacing=0.03,
        subplot_titles=[],
        column_widths=[1.0],
        row_heights=[1.0],
        specs=[[{'type': 'xy'}]],
        figure=fig)


    # Position traces in subplots
    for trace in trace_lst:
        trace.update({'xaxis': 'x', 'yaxis': 'y'})

    # Add traces, layout and frames to figure
    fig.add_traces(trace_lst)
    fig.update_layout({'barmode': 'overlay',
                       'legend': {'tracegroupgap': 0,
                                  'title_text': ACTIVITY}
                       })
    configure_axes(args, go.Bar, fig, orders)

    return fig


def contingency_events(df_acts=None, df_devs=None, con_tab=None, scale='linear', height=350,
                       act_order='alphabetical', dev_order='alphabetical', distributed=False
                       ) -> plotly.graph_objects.Figure:
    """
    """
    title = 'Activities vs device events'
    cbarlabel = 'count' if scale == 'linear' else 'log count'

    # get the list of cross tabulations per t_window
    if con_tab is not None:
        df_con = con_tab.copy()    # :pd.DataFrame
    else:
        df_con = contingency_table_events(df_devs, df_acts)

    df_con = df_con.set_index('device')

    act_order = activity_order_by(df_acts, rule=act_order)
    dev_order = device_order_by(df_devs, rule=dev_order)
    df_con = df_con[np.flip(act_order)]     # change order of rows
    df_con = df_con.reindex(dev_order)      # change order of columns

    # Extract values
    devs = df_con.index.values
    acts = df_con.columns.values
    vals = df_con.values
    z = vals if scale == 'linear' else np.log(vals)

    fig = go.Figure(data=go.Heatmap(
        z=z.T,
        x=devs,
        y=acts,
        customdata=vals.T,
        hovertemplate='"%{x}" during "%{y}" %{customdata} times<extra></extra>',
        colorscale='Viridis',
        hoverongaps=False))

    _set_compact_title(fig, title)

    fig.update_xaxes(tickangle=45)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    _style_colorbar(fig, cbarlabel)

    return fig


def contingency_states(df_acts=None, df_devs=None, con_tab=None, scale='linear', height=350,
                       act_order='alphabetical', dev_order='alphabetical', distributed=False
                       ) -> plotly.graph_objects.Figure:
    """
    """
    title = 'Activities vs device states'
    cbarlabel = 'seconds' if scale == 'linear' else 'log seconds'

    # get the list of cross tabulations per t_window
    if con_tab is not None:
        df_con = con_tab.copy()    # :pd.DataFrame
    else:
        df_con = contingency_table_states(df_devs, df_acts, distributed=distributed)

    # Determine device and activity labels order
    act_order = activity_order_by(df_acts, rule=act_order)
    df_con = df_con[np.flip(act_order)]     # change order of rows
    devs = df_con.index.values

    # Reorder and format devices
    dtypes = infer_dtypes(df_devs)
    dtypes.pop(NUM, None)
    new_devs, new_order = format_device_labels(devs, dtypes, boolean_state=True, categorical_state=True)
    devs = list(new_devs)
    vals = df_con.copy()\
        .reset_index()\
        .reindex(new_order)\
        .set_index(DEVICE)

    # Extract values
    acts = df_con.columns.values


    # Convert timedelta to nanoseconds
    z = vals.astype('timedelta64[ns]')/np.timedelta64(1, 'ns')
    z = z if scale == 'linear' else np.log(z)
    z = z.T

    # Create hoverdata, the duration as strings in (D,A,0) and full X names in (D,A,1)
    vals = vals.astype(str)
    tmp = np.tile(vals.index, (len(vals.columns), 1))
    cd = np.array([tmp, vals.values.T])
    cd = np.moveaxis(cd, 0, -1)


    # TODO hack, add whitespaces to similar categories to make them unique
    #      strings since otherwise plotly merges them into one category
    j = 0
    for i in range(len(devs)):
        if devs[i] == 'on':
            devs[i] = devs[i] + ' '*j
            j += 1

    fig = go.Figure(data=go.Heatmap(
        z=z,
        x=devs,
        y=acts,
        customdata=cd,
        hovertemplate='"%{customdata[0]}" during "%{y}" for %{customdata[1]}<extra></extra>',
        colorscale='Viridis',
        hoverongaps=False))

    _set_compact_title(fig, title)

    fig.update_xaxes(tickangle=45)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    _style_colorbar(fig, cbarlabel)

    return fig



import sys
import pathlib

from sklearn.neighbors import KernelDensity
from sklearn.svm import SVC

working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest
from sklearn.ensemble import RandomForestClassifier
from pyadlml.dataset import set_data_home, load_act_assist
from pyadlml.pipeline import Pipeline
from pyadlml.preprocessing import StateVectorEncoder, LabelEncoder, DropTimeIndex
from pyadlml.model_selection import GridSearchCV
from pyadlml.pipeline import Pipeline
from pyadlml.preprocessing import TestSubset, TrainSubset
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier

SUBJECT_ADMIN_NAME = 'admin'

class TestGridSearchCV(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/partial_dataset'
        self.data = load_act_assist(dataset_dir, subjects=[SUBJECT_ADMIN_NAME])
        self.data.df_activities = self.data.df_activities_admin
        self.iris = load_iris()

    def test_score_samples(self):
        X = self.iris['data']
        y = self.iris['target']
        param_grid = {'classifier__n_estimators': [100, 50]}
        steps = [('classifier', RandomForestClassifier(random_state=42))]
        pipe = Pipeline(steps).train()
        tmp = GridSearchCV(
            estimator=pipe,
            param_grid=param_grid,
            scoring='accuracy',
            verbose=1,
        )
        tmp.fit(X,y)



if __name__ == '__main__':
    unittest.main()


import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pandas as pd

from pyadlml.constants import ACTIVITY, TIME, START_TIME, END_TIME
from pyadlml.dataset.stats.activities import activities_duration_dist, activity_duration, \
    activities_transitions, activities_count, activity_duration, activities_dist, activity_order_by_duration
from pyadlml.dataset._core.activities import add_other_activity
from .util import func_formatter_seconds2time_log, ridgeline, \
    func_formatter_seconds2time, heatmap, annotate_heatmap, heatmap_square, savefig, \
    _num_bars_2_figsize, _num_boxes_2_figsize, \
    _num_items_2_heatmap_square_figsize, _num_items_2_ridge_figsize, \
    _num_items_2_ridge_ylimit, xaxis_format_one_day, save_fig, get_qualitative_cmap, plot_grid, xaxis_format_time2, \
    map_time_to_numeric
from pyadlml.util import get_sequential_color, get_secondary_color, get_primary_color, get_diverging_color
from ..util import activity_order_by


@save_fig
def count(df_acts=None, df_ac=None, scale="linear", other=False,
          figsize=None, color=None, file_path=None):
    """
    Plot a bar chart displaying how often activities are occurring.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    other : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    scale : {"log", "linear"}, default: linear
        The axis scale type to apply.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_activities_count
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> data = fetch_amsterdam()
    >>> plot_activities_count(data.df_activities, idle=True)

    .. image:: ../_static/images/plots/act_bar_cnt.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing 
    """
    assert not (df_acts is None and df_ac is None)
    assert scale in ['linear', 'log']

    title ='Activity occurrences'
    col_label = 'occurrence'
    x_label = 'counts'
    color = (get_primary_color() if color is None else color)
        
    # create statistics if the don't exists
    if df_ac is None:
        df_acts = df_acts.copy()
        if other:
            df_acts = add_other_activity(df_acts)
        df = activities_count(df_acts)
    else:
        df = df_ac
    
    # prepare dataframe for plotting
    df.reset_index(level=0, inplace=True)
    df = df.sort_values(by=[col_label], axis=0)
    
    # define plot modalities
    num_act = len(df)
    figsize = (_num_bars_2_figsize(num_act) if figsize is None else figsize)
    
    # create plot
    fig, ax = plt.subplots(figsize=figsize)
    plt.title(title)
    plt.xlabel(x_label)
    ax.barh(df[ACTIVITY], df[col_label], color=color)
    
    if scale == 'log':
        ax.set_xscale('log')

    return fig


@save_fig
def boxplot(df_acts, scale='linear', other=False, figsize=None,
            order='alphabetical', file_path=None):
    """
    Plot a boxplot for activity durations.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    other : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *other* or be ignored.
    order : str, default='alphabetical'
        The order in which the activities are displayed
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_boxplot
    >>> plot_activity_boxplot(data.df_activities)

    .. image:: ../_static/images/plots/act_bp.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing
    """
    assert scale in ['linear', 'log']
    
    title = 'Activity durations'
    xlabel = 'Seconds'

    if other:
        df_acts = add_other_activity(df_acts)

    act_order = np.flip(activity_order_by(df_acts, rule=order))
    df = activities_duration_dist(df_acts)

    # Select data for each device
    df['seconds'] = df['minutes']*60

    num_act = len(act_order)
    figsize = (_num_bars_2_figsize(num_act) if figsize is None else figsize)

    dat = []
    for activity in act_order:
        df_activity = df[df[ACTIVITY] == activity]
        dat.append(df_activity['seconds'])
    
    # Plot boxsplot
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(dat, vert=False)
    ax.set_title(title)
    ax.set_yticklabels(act_order, ha='right')
    ax.set_xlabel(xlabel)

    if scale == 'log':
        ax.set_xscale('log')

    # Create secondary axis with time format 1s, 1m, 1d
    ax_top = ax.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_seconds2time))

    return fig


@save_fig
def duration(df_acts=None, df_dur=None, scale='linear', idle=False,
             figsize=None, color=None, file_path=None):
    """
    Plots the cumulative duration for each activity in a bar plot.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import duration
    >>> duration(data.df_activities)

    .. image:: ../_static/images/plots/act_bar_dur.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing
    """
    assert scale in ['linear', 'log']
    assert not (df_acts is None and df_dur is None)

    title = 'Cummulative activity durations'
    xlabel = 'seconds'
    freq = 'seconds'
    color = (get_primary_color() if color is None else color)

    if df_dur is None:
        if idle:
            df_acts = add_other_activity(df_acts.copy())
        df = activity_duration(df_acts, time_unit=freq)
    else:
        df = df_dur
    df = df.sort_values(by=[freq], axis=0)

    num_act = len(df)
    figsize = (_num_bars_2_figsize(num_act) if figsize is None else figsize)
    
    # plot
    fig, ax = plt.subplots(figsize=figsize)
    plt.title(title)
    plt.xlabel(xlabel)
    ax.barh(df[ACTIVITY], df['seconds'], color=color)
    if scale == 'log':
        ax.set_xscale('log')
        
        
    # create secondary axis with time format 1s, 1m, 1d
    ax_top = ax.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    ax_top.set_xlabel('time')
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_seconds2time))

    return fig


@save_fig
def transitions(df_acts=None, df_trans=None, scale="linear",
                figsize=None, idle=False, numbers=True, grid=True,
                cmap=None, file_path=None):
    """
    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    df_trans : pd.DataFrame
        A precomputed transition table. If the *df_trans* parameter is given, parameters
        *df_acts* and *lst_acts* are ignored. The transition table can be computed
        in :ref:`stats <stats_acts_trans>`.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    numbers : bool, default: True
        Whether to display numbers inside the heatmaps fields or not.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    cmap : str or Colormap, optional
        The Colormap instance or registered colormap name used to map scalar
        data to colors. This parameter is ignored for RGB(A) data.
        Defaults 'viridis'.
    grid : bool, default: True
        determines whether to display a white grid, seperating the fields or not.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import transitions
    >>> transitions(data.df_activities)

    .. image:: ../_static/images/plots/act_hm_trans.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center


    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert scale in ['linear', 'log'], 'scale has to be either of type None or log'
    assert not (df_acts is None and df_trans is None)

    title = 'Activity transitions'
    z_label = 'count'

    if df_trans is None:
        df_acts = add_other_activity(df_acts) if idle else df_acts
        df = activities_transitions(df_acts)
    else:
        df = df_trans

    # get the list of cross tabulations per t_window
    act_lst = list(df.columns)

    num_act = len(act_lst)
    figsize = (_num_items_2_heatmap_square_figsize(num_act) if figsize is None else figsize)
    cmap = (get_sequential_color() if cmap is None else cmap)

    x_labels = act_lst
    y_labels = act_lst
    values = df.values
    

    log = True if scale == 'log' else False
    valfmt = '{x:.0f}'
        
     # begin plotting
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(values, y_labels, x_labels, log=log, cmap=cmap, ax=ax, cbarlabel=z_label, grid=grid)
    if numbers:
        texts = annotate_heatmap(im, textcolors=("white", "black"),log=log, valfmt=valfmt)
    ax.set_title(title)
    
    return fig


@save_fig
def density(df_acts=None, df_act_dist=None, idle=False, dt=None,
            n=1000, ylim_upper=None, color=None, figsize=None, file_path=None):
    """
    Plots the activity density distribution over one day.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    df_act_dist : pd.DataFrame, optional
        A precomputed activity density distribution. If the *df_trans* parameter is given, parameters
        *df_acts* and *lst_acts* are ignored. The transition table can be computed
        in :ref:`stats <stats_acts_trans>`.
    n : int, default=1000
        The number of monte-carlo samples to draw.
    ylim_upper: float, optional
        The offset from the top of the plot to the first ridge_line. Set this if
        the automatically determined value is not satisfying.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_density
    >>> plot_activity_density(data.df_activities)

    .. image:: ../_static/images/plots/act_ridge_line.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified otherwise nothing.
    """
    assert not (df_acts is None and df_act_dist is None)

    title = 'Activity density over one day'
    xlabel = 'Time'
    color = (get_primary_color() if color is None else color)


    if df_act_dist is None:
        if idle:
            df_acts = add_other_activity(df_acts)
        df = activities_dist(df_acts.copy(), n=n, dt=dt, relative=True)
        if df.empty:
            raise ValueError("no activity was recorded and no activity list was given.")
    else:
        df = df_act_dist

    if dt is None:
        # plot ridgeline
        grouped = [(activity, df.loc[df[ACTIVITY] == activity, TIME].sort_values().values)
                   for activity in df[ACTIVITY].unique()]
        acts, data = zip(*grouped)
        num_act = len(list(acts))

        # infer visual properties
        figsize = (_num_items_2_ridge_figsize(num_act) if figsize is None else figsize)
        ylim_upper = (_num_items_2_ridge_ylimit(num_act) if ylim_upper is None else ylim_upper)

        # plot the ridgeline
        fig, ax = plt.subplots(figsize=figsize)
        ridgeline(data, labels=acts, overlap=.85, fill=color, n_points=100, dist_scale=0.13)
        plt.title(title)

        plt.gca().spines['left'].set_visible(False)
        plt.gca().spines['right'].set_visible(False)
        plt.gca().spines['top'].set_visible(False)
        plt.ylim((0, ylim_upper))
        plt.xlabel(xlabel)

        xaxis_format_one_day(ax, 0, 60*60*24)
        fig.autofmt_xdate(rotation=45)

        plt.grid(zorder=0)
    else:
        # plot heatmap
        x_labels = list(df[TIME])
        activities = df.columns[1:]
        dat = df.iloc[:, 1:].values.T

        cmap = get_sequential_color()
        cbarlabel = 'counts'

        # begin plotting
        fig, ax = plt.subplots(figsize=figsize)
        im, cbar = heatmap(dat, activities, x_labels, ax=ax, cmap=cmap, cbarlabel=cbarlabel)
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        xaxis_format_one_day(ax, *ax.get_xlim())

        ax.set_aspect(aspect='auto')

    return fig


@save_fig
def correction(df_act_pre, df_act_post, grid=True, figsize=(10, 5), file_path=None):
    """ Plots the activities before and after the automatic correction. Refer to the user guide: TODO


    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    df_act_dist : pd.DataFrame, optional
        A precomputed activity density distribution. If the *df_trans* parameter is given, parameters
        *df_acts* and *lst_acts* are ignored. The transition table can be computed
        in :ref:`stats <stats_acts_trans>`.
    n : int, default=1000
        The number of monte-carlo samples to draw.
    ylim_upper: float, optional
        The offset from the top of the plot to the first ridge_line. Set this if
        the automatically determined value is not satisfying.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> from pyadlml.plot import plot_activity_correction

    >>> corr = fetch_amsterdam().correction_activities
    >>> plot_activity_correction(corr[0][0], corr[0][1])

    .. image:: ../_static/images/plots/act_ridge_line.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified otherwise nothing.

    """
    title = 'Activity corrections'
    activity_y_label = 'Activities'
    right_title = 'Post-correction'
    left_title = 'Prae-correction'

    df_act_post = df_act_post.sort_values(by=START_TIME)
    df_act_pre = df_act_pre.sort_values(by=START_TIME)

    start_time = min(df_act_pre[START_TIME].iloc[0], df_act_post[START_TIME].iloc[0])
    end_time = max(df_act_pre[END_TIME].iloc[-1], df_act_post[END_TIME].iloc[-1])

    df_act_post = df_act_post.sort_values(by=START_TIME, ascending=False).reset_index(drop=True)
    df_act_pre = df_act_pre.sort_values(by=START_TIME, ascending=False).reset_index(drop=True)

    acts = activity_order_by_duration(df_act_pre)

    df_act_pre['num_st'], _, _ = map_time_to_numeric(df_act_pre[START_TIME], start_time, end_time)
    df_act_pre['num_et'], _, _ = map_time_to_numeric(df_act_pre[END_TIME], start_time, end_time)
    df_act_pre['diff'] = df_act_pre['num_et'] - df_act_pre['num_st']

    df_act_post['num_st'], _, _ = map_time_to_numeric(df_act_post[START_TIME], start_time, end_time)
    df_act_post['num_et'], _, _ = map_time_to_numeric(df_act_post[END_TIME], start_time, end_time)
    df_act_post['diff'] = df_act_post['num_et'] - df_act_post['num_st']


    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)
    tab = get_qualitative_cmap(len(acts))

    if grid:
        plot_grid(fig, ax1, start_time, end_time)
        plot_grid(fig, ax2, start_time, end_time)

    # create a column with the color for each department
    def color(row):
        return tab(np.where(np.array(acts) == row[ACTIVITY])[0][0])

    df_act_pre['color'] = df_act_pre.apply(color, axis=1)
    df_act_post['color'] = df_act_post.apply(color, axis=1)

    ax1.barh(df_act_pre.index, df_act_pre['diff'], left=df_act_pre['num_st'],
             color=df_act_pre['color'], label=df_act_pre[ACTIVITY])

    ax2.barh(df_act_post.index, df_act_post['diff'], left=df_act_post['num_st'],
             color=df_act_post['color'], label=df_act_post[ACTIVITY])

    ax1.set_xlim(-0.05, 1.05)
    ax2.set_xlim(-0.05, 1.05)

    from matplotlib.patches import Patch
    legend_elements = [Patch(facecolor=tab(i), label=act) for i, act in enumerate(acts)]

    ax1.set_title(left_title)
    ax2.set_title(right_title)
    ax1.set_yticks([])
    ax2.set_yticks([])
    ax2.legend(handles=legend_elements, bbox_to_anchor=(1, 1), loc="upper left")
    xaxis_format_time2(fig, ax1, start_time, end_time)
    xaxis_format_time2(fig, ax2, start_time, end_time)

    return fig


import matplotlib.pyplot as plt
import matplotlib
import numpy as np
import pandas as pd
from mpl_toolkits.axes_grid1 import make_axes_locatable

from pyadlml.constants import ACTIVITY, DEVICE, VALUE, START_TIME, END_TIME, TIME, BOOL, CAT, NUM
from pyadlml.dataset._core.devices import contains_non_binary, split_devices_binary, device_events_to_states
from pyadlml.dataset._core.activities import add_other_activity
from pyadlml.dataset.stats import contingency_states as stat_cont_st, contingency_events as stat_cont_ev, cross_correlogram as stat_cc
from .devices import _plot_device_states
from .util import func_formatter_seconds2time, save_fig, \
    heatmap_square, heatmap, annotate_heatmap, heatmap_contingency, map_time_to_numeric, \
    get_qualitative_cmap, plot_cc, create_todo, xaxis_format_time, format_device_labels, xaxis_format_time2, \
    dev_raster_data_gen, format_device_labels, plot_grid
from pyadlml.dataset.util import select_timespan, infer_dtypes, str_to_timestamp

from pyadlml.dataset.stats.activities import activity_order_by_duration

DEV_CON_HM = {(8, 12): (8, 7), (8, 14): (8, 8), (8, 3): (12, 10), (26, 20): (10, 10), (11, 17): (10, 10)}
DEV_CON_01_HM = {(8, 14): (12, 10), (11, 24): (12, 6), (11, 34): (16, 12)}
DEV_CON_01_HM_WT = {(10, 24): (14, 7), (7, 28): (16, 10)}


@save_fig
def cross_correlogram(df_devices, df_actvities, corr_data=[None, None, None, None], binsize='1s', maxlag='2m',
                      axis='off', figsize=(5, 5), file_path=""):
    #Plot cross-correlograms of all pairs.
    #   plotCCG(ccg,bins) plots a matrix of cross(auto)-correlograms for
    #   all pairs of clusters. Inputs are:
    #       ccg     array of cross correlograms           #bins x #clusters x #clusters
    #       bins    array with bin timings                #nbins x 0

    title = 'Cross-correlogram'

    if corr_data[0] is None or corr_data[1] is None:
        ccg, bins = stat_cc(df_devices, df_actvities, binsize=binsize, maxlag=maxlag)
        devices = df_devices[DEVICE].unique()
        activities = df_actvities[ACTIVITY].unique()
    else:
        ccg = corr_data[0]
        bins = corr_data[1]
        devices = corr_data[2]
        activities = corr_data[3]

    return plot_cc(ccg, bins, title=title, x_label=activities, y_label=devices, axis=axis, figsize=figsize)


@save_fig
def contingency_events(df_devs=None, df_acts=None, df_con_tab=None, idle=False, per_state=False, \
                       scale=None, numbers=True, figsize=None, file_path=""):
    """ Computes a table where the device triggers are counted per activity

    Parameters
    ----------
    df_devs : pd.DataFrame, optional
        recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`. If the parameter *df_devs* is not set,
        the parameter *df_con_tab* has to be set.
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`. If the parameter *df_acts* is not set,
        the parameter *df_con_tab* has to be set.
    df_con_tab : pd.DataFrame, optional
        A precomputed contingency table. If the *df_con_tab* parameter is given, parameters
        *df_acts* and *df_devs* are ignored. The contingency table can be computed
        in :ref:`stats <stats_dna_con_dur>`.
    per_state : bool, default=False
        If set events are
    figsize : (float, float), optional
        width, height in inches. If not provided, the figsize is inferred by automatically.
    scale : {"log", "linear"}, default: 'log'
        The axis scale type to apply.
    numbers : bool, default: True
        Whether to display numbers inside the heatmaps fields or not.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_contingency_events
    >>> plot_contingency_events(data.df_activities, data.df_activities)

    .. image:: ../_static/images/plots/cont_hm_trigger.png
       :height: 300px
       :width: 500 px
       :scale: 100 %
       :alt: alternate text
       :align: center

    >>> contingency_events(data.df_devices, data.df_activities, per_state=True)

    .. image:: ../_static/images/plots/cont_hm_trigger.png
       :height: 300px
       :width: 500 px
       :scale: 100 %
       :alt: alternate text
       :align: center

    Returns
    -------
    fig : Figure or None
        If the parameter file_path is specified, the method return None rather than a matplotlib figure.

    """
    assert (df_devs is not None and df_acts is not None) or df_con_tab is not None

    title = 'Events'
    cbarlabel = 'counts'
    textcolors = ("white", "black")
    # if log than let automatically infer else
    log = (scale == 'log')
    valfmt = (None if log else "{x:.0f}")

    if df_con_tab is None:
        ct = stat_cont_ev(df_devs, df_acts, idle=idle, per_state=per_state)
    else:
        ct = df_con_tab.copy()

    vals = ct.iloc[:, 1:].values.T
    acts = ct.columns[1:].values
    devs = ct[DEVICE].values

    # format labels
    dtypes = infer_dtypes(df_devs)
    if per_state:
        new_devs, new_order = format_device_labels(devs, dtypes, boolean_state=True, categorical_state=True)
    else:
        new_devs, new_order = format_device_labels(devs, dtypes)

    devs = list(new_devs)
    vals = vals[:, new_order]


    fig = heatmap_contingency(acts, devs, vals, title, cbarlabel, valfmt=valfmt,
                              textcolors=textcolors, z_scale=scale, numbers=numbers, figsize=figsize)

    return fig



@save_fig
def plot_contingency_states(df_devs: pd.DataFrame = None, df_acts: pd.DataFrame = None, df_con_tab: pd.DataFrame = None,
                            figsize: tuple = None, z_scale: str = 'log', idle: bool = False, numbers: bool = True,
                            file_path: str = "") -> object:
    """
    Plots a heatmap the device on and off intervals are measured against
    the activities

    Parameters
    ----------
    df_devs : pd.DataFrame, optional
        recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`. If the parameter *df_devs* is not set,
        the parameter *df_con_tab* has to be set.
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`. If the parameter *df_acts* is not set,
        the parameter *df_con_tab* has to be set.
    df_con_tab : pd.DataFrame, optional
        A precomputed contingency table. If the *df_con_tab* parameter is given, parameters
        *df_acts* and *df_devs* are ignored. The contingency table can be computed
        in :ref:`stats <stats_dna_con_dur>`.
    figsize : (float, float), optional
        width, height in inches. If not provided, the figsize is inferred by automatically.
    z_scale : {"log", "linear"}, default: 'log'
        The axis scale type to apply.
    numbers : bool, default: True
        Whether to display numbers inside the heatmaps fields or not.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> from pyadlml.plot import contingency_states
    >>> data = fetch_amsterdam()
    >>> fig = contingency_states(data.df_devices, data.df_activities)
    >>> fig.show()

    .. image:: ../_static/images/plots/cont_hm_duration.png
       :height: 300px
       :width: 800 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    fig : Figure or None
        If the parameter file_path is specified, the method returns None rather than a matplotlib figure.

    """

    assert (df_devs is not None and df_acts is not None) or df_con_tab is not None

    title = 'Activity vs. device states'
    cbarlabel = 'time in seconds'

    if df_con_tab is None:
        if idle:
            df_acts = add_other_activity(df_acts.copy())
        df_con = stat_cont_st(df_devs, df_acts)
    else:
        df_con = df_con_tab

    # convert time (ns) to seconds
    df_con = df_con.astype(int)/1000000000

    devs = df_con.index.values
    vals = df_con.values.T
    acts = df_con.columns.values

    # sort values and format device state labels
    if df_devs is not None:
        dtypes = infer_dtypes(df_devs)
    else:
        dtypes = {BOOL: [d.split(':')[0] for d in devs if (':on' in d) or (':off' in d)],
                  CAT: [d.split(':')[0] for d in devs if not (':on' in d) and not (':off' in d)],
                  NUM: []}
    dtypes.pop(NUM, None)
    new_devs, new_order = format_device_labels(devs, dtypes, boolean_state=True, categorical_state=True)
    devs = list(new_devs)
    vals = vals[:, new_order]


    valfmt = matplotlib.ticker.FuncFormatter(lambda x, p: func_formatter_seconds2time(x, p))
    
    return heatmap_contingency(acts, devs, vals, title, cbarlabel,
             valfmt=valfmt, figsize=figsize, z_scale=z_scale,
             numbers=numbers)


@save_fig
def activities_and_device_events(df_devices, df_activities, start_time=None, end_time=None,
                                 figsize=(20, 8), grid=False, file_path=""):
    """ Plots the devices events as raster combined with the activites over a timespan

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe
    df_activities : pd.DataFrame
        An activity dataframe
    start_time : str or None, default=None
        When set, a string representing the time where the plots starts
    end_time : str or None, default=None
        When set, a string representing the time where the plots ends
    figsize : tuple, default=(20,8)
        The plots figure given in (width, height)
    file_path : str, default=""
        When set the file

    Returns
    ------
    fig or None
        Either a figure or nothing depending on if file_path is set.
    """
    title='Activities vs. device events'
    y_label = 'Devices'
    event_color = 'grey'

    activity_y_label = 'Activities'
    act_offset_space = 0.25
    act_bar_height = 1


    df_acts = df_activities.copy()
    df_devs = df_devices.copy()\
                        .sort_values(by=TIME)\
                        .reset_index(drop=True)

    devs = df_devs[DEVICE].unique()
    acts = activity_order_by_duration(df_acts)

    start_time = str_to_timestamp(start_time) if (start_time is not None) else None
    end_time = str_to_timestamp(end_time) if (end_time is not None) else None
    df_devs, df_acts = select_timespan(df_devs, df_acts, start_time, end_time, clip_activities=True)
    if start_time is None:
        start_time = min(df_devs[TIME].iloc[0], df_acts[START_TIME].iloc[0])
    if end_time is None:
        end_time = max(df_devs[TIME].iloc[-1], df_acts[END_TIME].iloc[-1])


    df_devs['num_time'], start_time, end_time = map_time_to_numeric(df_devs[TIME], start_time, end_time)
    data_lst = dev_raster_data_gen(df_devs, devs)

    fig, ax = plt.subplots(figsize=figsize)

    tab = get_qualitative_cmap(len(acts))
    plot_activity_bar(ax, df_acts, acts, start_time, end_time, act_bar_height,
                      len(devs)+act_offset_space, tab(np.arange(0, len(acts))))

    ax.eventplot(data_lst, linelengths=[0.2]*len(devs), colors=event_color)


    ax.set_title(title)
    yticks = np.arange(len(devs)+1, dtype=np.float32)
    yticks[-1] += act_offset_space + act_bar_height/2    # correct offset for activities
    ax.set_yticks(yticks)
    ax.set_yticklabels(devs.tolist() + [activity_y_label])
    ax.set_ylabel(y_label)

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles, labels, bbox_to_anchor=(1, 1), loc="upper left")
    ax.set_xlim([0, 1])

    xaxis_format_time2(fig, ax, start_time, end_time)

    if grid:
        plot_grid(fig, ax, start_time, end_time)

    return fig





@save_fig
def activities_and_device_states(df_devices: pd.DataFrame, df_activities: pd.DataFrame,
                                 start_time: str = None, end_time: str = None,
                                 figsize=(20, 8), grid=False, file_path=""):
    """ Plots devices events as raster combined with activities over time

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe
    df_activities : pd.DataFrame
        An activity dataframe
    start_time : str or None, default=None
        When set, a string representing the time where the plots starts
    end_time : str or None, default=None
        When set, a string representing the time where the plots ends
    figsize : tuple, default=(20,8)
        The plots figure given in (width, height)
    file_path : str, default=""
        When set the file

    Returns
    ------
    fig or None
        Either a figure or nothing depending on if file_path is set.
    """

    ylabel = 'Devices'
    title = 'Device states'
    binary_on_label = 'on'
    binary_off_label = 'off'
    activity_y_label = 'Activities'

    color_dev_on = 'seagreen'
    color_dev_off = 'lightgrey'
    color_dev_num = 'blue'

    act_offset_space = 0.25
    act_bar_height = 1

    df_acts = df_activities.copy()
    df_devs = df_devices.copy()\
                        .sort_values(by=TIME)\
                        .reset_index(drop=True)


    start_time = str_to_timestamp(start_time) if (start_time is not None) else None
    end_time = str_to_timestamp(end_time) if (end_time is not None) else None
    df_devs, df_acts = select_timespan(df_devices=df_devs, df_activities=df_acts,
                                       start_time=start_time, end_time=end_time,
                                       clip_activities=True)


    if start_time is None:
        start_time = min(df_devs[TIME].iloc[0], df_acts[START_TIME].iloc[0])
    if end_time is None:
        end_time = max(df_devs[TIME].iloc[-1], df_acts[END_TIME].iloc[-1])

    devs = df_devs[DEVICE].unique()
    acts = activity_order_by_duration(df_acts)

    fig, ax = plt.subplots(figsize=figsize)
    tab = get_qualitative_cmap(len(acts))

    if grid:
        plot_grid(fig, ax, start_time, end_time)

    j = _plot_device_states(ax, df_devs, devs, start_time, end_time, color_dev_on, color_dev_off,
                           color_dev_num, binary_off_label, binary_on_label, return_nr_categories_used=True)

    plot_activity_bar(ax, df_acts, acts, start_time, end_time, act_bar_height,
                      len(devs)+act_offset_space,
                      tab(np.arange(j, len(acts)+j))
    )

    ax.set_title(title)
    yticks = np.arange(len(devs)+1, dtype=np.float32)
    yticks[-1] += act_offset_space + act_bar_height/2    # correct offset for activities

    ax.set_yticks(yticks)
    ax.set_yticklabels(devs.tolist() + [activity_y_label])
    ax.set_ylabel(ylabel)
    ax.set_xlim([0, 1])

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles, labels, bbox_to_anchor=(1, 1), loc="upper left")
    xaxis_format_time2(fig, ax, start_time, end_time)

    return fig

def plot_activity_bar(ax, df_acts: pd.DataFrame, acts: list, start_time, end_time, act_bar_height, y_pos, c_map):
    """ Plot a horizontal activity event representation.
    Parameters
    ----------
    c_map : np.ndarray
        Array of colors where each color corresponds to one activity
    """
    df_acts['num_st'], _,_ = map_time_to_numeric(df_acts[START_TIME], start_time, end_time)
    df_acts['num_et'], _,_ = map_time_to_numeric(df_acts[END_TIME], start_time, end_time)
    df_acts['diff'] = df_acts['num_et'] - df_acts['num_st']

    # create list of lists, where each list corresponds to an activity with tuples of start_time and time_length
    xranges = []
    for i, act in enumerate(acts):
        xranges.append(df_acts.loc[df_acts[ACTIVITY] == act, ['num_st', 'diff']].values.tolist())

    for i in range(len(acts)):
        ax.broken_barh(xranges[i], (y_pos, act_bar_height), facecolors=c_map[i], label=acts[i])



def plot_events_over_act_hist(df_acts: pd.DataFrame, df_devs: pd.DataFrame, activity: str, device: str, dt: str = '1s') -> plt.Figure:
    """ Plot a histogram when events happen centered around an activity

    Parameters
    ----------
    df_acts : pd.DataFrame
        An activity dataframe with columns: 'start_time', 'end_time' and 'activity'


    Returns
    -------
    Figure or None

    """
    pass


import functools
import plotly.figure_factory as ff
from .util import _style_colorbar
import plotly.express as px
import pandas as pd
import numpy as np
import plotly.express as px
import plotly.graph_objects
import plotly.graph_objects as go
from plotly.colors import n_colors
from plotly.subplots import make_subplots

from pyadlml.constants import START_TIME, ACTIVITY, TIME, END_TIME, STRFTIME_DATE, PRIMARY_COLOR, \
                              SECONDARY_COLOR
from pyadlml.dataset.stats.activities import activities_duration_dist, \
    activities_count, activity_duration, activities_dist, \
    activities_transitions
from pyadlml.dataset.util import get_sorted_index, check_scale, check_order, activity_order_by


def density(df_act: pd.DataFrame = None, df_density: pd.DataFrame = None,
            dt=None, n=1000, height=350, order='alphabetical', show_meanline=True) -> plotly.graph_objects.Figure:
    """
    https://plotly.com/python/violin/

    for one day plot the activity distribution over the day
    - sample uniform from each interval   
    """

    title = 'Density'

    if df_act is None:
        df = df_density
    else:
        df = activities_dist(df_act.copy(), n=n, dt=dt)

    acts = np.flip(activity_order_by(df, rule=order))

    #colors = n_colors('rgb(5, 200, 200)', 'rgb(200, 10, 10)', len(acts), colortype='rgb')
    colors = [PRIMARY_COLOR]*len(acts)
    fig = go.Figure()

    for activity, color in zip(acts, colors):
        data_line = df[df[ACTIVITY] == activity][TIME]
        data_line = pd.to_datetime(data_line)
        fig.add_trace(go.Violin(x=data_line, line_color=color, name=activity,
                                hoverinfo='skip'))

    # Set custom x-axis labels
    fig.update_xaxes(tickformat="%H:%M", tickangle=-45,     # nticks=24,
                     title_text='Time', range=[
                        data_line.iloc[0].floor('D'),
                        data_line.iloc[0].ceil('D')
                     ])
    fig.update_traces(orientation='h', side='positive', width=1, points=False)
    fig.update_layout(xaxis_showgrid=True, xaxis_zeroline=False, showlegend=False,
                      height=height, margin=dict(l=0, r=0, b=0, t=30, pad=0))
    _set_compact_title(fig, title=title)
    if show_meanline:
        fig.update_traces(meanline_visible=True)

    return fig

@check_scale
def _scale_xaxis(fig, scale, values) -> None:
    if scale == 'log':
        r_min = np.floor(np.log10(values.min()))
        r_max = np.ceil(np.log10(values.max()))
        fig.update_xaxes(type="log", range=[r_min, r_max])  # log range: 10^r_min=1, 10^r_max=100000

@check_scale
def violin_duration(df_act, x_scale='linear', no_title=False, height=350) -> plotly.graph_objects.Figure:
    df = activities_duration_dist(df_act)
    # Add column for hover display of datapoints later
    df[START_TIME] = df_act[START_TIME].dt.strftime('%c')
    df[END_TIME] = df_act[END_TIME].dt.strftime('%c')

    fig = px.violin(df, y='activity', x='minutes', box=False, orientation='h', points='outliers',
                    hover_data=[START_TIME, END_TIME, 'minutes'])

    if not no_title:
        fig.update_layout(title='Activity distribution')
    _scale_xaxis(fig, x_scale, df['minutes'])
    fig.update_yaxes(title=None, visible=False, showticklabels=False)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=0, pad=0), height=height)
    return fig

def _set_compact_title(fig: plotly.graph_objects.Figure, title: str) -> None:
    fig.update_layout(title=dict(
        text=title,
        yanchor='top',
        xanchor='left',
        pad={'t': 2, 'l': -20}
    ))

@check_scale
def boxplot_duration(df_act, scale='linear', height=350, order='alphabetical') -> plotly.graph_objects.Figure:
    """ Plot a boxplot of activity durations (mean) max min
    """
    title = 'Duration'

    df = activities_duration_dist(df_act)
    act_order = activity_order_by(df, rule=order)

    # Add column for hover display of datapoints later
    df[START_TIME] = df_act[START_TIME]
    df[END_TIME] = df_act[END_TIME]
    df['total_time'] = df['total_time'].apply(str)

    fig = px.box(df, y="activity", x='minutes', orientation='h',
                 category_orders={'activity': act_order},
                 color_discrete_sequence=[PRIMARY_COLOR],
                 notched=False, points='all',
                 hover_data=[START_TIME, END_TIME, 'minutes', 'total_time'])

    _set_compact_title(fig, title)
    _scale_xaxis(fig, scale, df['minutes'])

    fig.update_yaxes(title=None, visible=False, showticklabels=False)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    hover_template = 'Minutes=%{x}=%{customdata[2]}<br>Activity=%{y}<br>'\
                     + 'Start_time=%{customdata[0]|' + STRFTIME_DATE + '}<br>'\
                     + 'End_time=%{customdata[1]|' + STRFTIME_DATE + '}<extra></extra>'
    fig.data[0].hovertemplate = hover_template
    return fig



@check_scale
def bar_cum(df_act, scale='linear', height=350, order='alphabetical', no_title=False) -> plotly.graph_objects.Figure:
    """ Plots the cumulated activities durations in a histogram for each activity
    """

    title = 'Duration'
    x_label = 'minutes' if scale == 'linear' else 'log minutes'
    df = activity_duration(df_act.copy())
    df = df.loc[get_sorted_index(df, order), :]

    # Create hoverstrings to represent the timedelta
    format_td = lambda x: ':'.join(str(x).split(':')[:-1]) \
                         +':' + str(x).split(':')[-1][:5]
    df['td'] = pd.to_timedelta(df['minutes'], unit='m')\
                 .apply(format_td)

    fig = go.Figure()
    fig.add_trace(go.Bar(x=df['minutes'],
                         customdata=df['td'],
                         hovertemplate="%{y} with %{customdata}<extra></extra>",
                         y=df[ACTIVITY],
                         marker={'color': PRIMARY_COLOR},
                         orientation='h',
                         name='duration'))

    _set_compact_title(fig, title)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    if isinstance(order, list) or isinstance(order, np.ndarray):
        fig.update_yaxes(title=None, categoryorder='array', categoryarray=np.flip(order))
    else:
        fig.update_yaxes(title=None)
    fig.update_xaxes(title=x_label)
    if scale == 'log':
        fig.update_xaxes(type='log')
    return fig

@check_scale
def bar_count(df_act, scale='linear', height=350, no_title=False, order='count') -> plotly.graph_objects.Figure:
    """ Plots the activities durations against each other
    """
    title ='Counts'
    col_label = 'occurrence'
    xlabel = 'log count' if scale == 'log' else 'count'

    df = activities_count(df_act.copy())
    act_order = activity_order_by(df_act, rule=order)


    fig = px.bar(df, y='activity',
                 category_orders={'activity': act_order},
                 color_discrete_sequence=[PRIMARY_COLOR],
                 x='occurrence',
                 orientation='h')

    _set_compact_title(fig, title)

    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    fig.update_yaxes(title=None)
    fig.update_xaxes(title=xlabel)
    if scale == 'log':
        fig.update_xaxes(type='log')
    return fig

@check_scale
def heatmap_transitions(df_act, scale='linear', height=350, order='alphabetical') -> plotly.graph_objects.Figure:
    """
    """
    title = 'Transitions'
    cbarlabel = 'count' if scale == 'linear' else 'log count'

    # get the list of cross tabulations per t_window
    df = activities_transitions(df_act)
    act_lst = list(df.columns)


    act_order = activity_order_by(df_act, rule=order)
    df = df[act_order]          # change order of rows
    df = df.reindex(act_order)  # change order of columns
    z = df.values if scale == 'linear' else np.log(df.values)

    fig = go.Figure(data=go.Heatmap(
        z=z.T,
        x=act_order,
        y=act_order,
        hovertemplate='"%{x}" then "%{y}" %{z} times<extra></extra>',
        colorscale='Viridis',
        hoverongaps=False))

    if scale == 'log':
        fig.data[0].hovertemplate = '"%{x}" then "%{y}" %{customdata} times<extra></extra>'
        cd = np.array([df.values.T])
        fig['data'][0]['customdata'] = np.moveaxis(cd, 0, -1)

    _set_compact_title(fig, title)

    fig.update_xaxes(tickangle=45)
    fig.update_yaxes(visible=False, showticklabels=False)
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0), height=height)
    _style_colorbar(fig, cbarlabel)
    return fig



import os
from pathlib import Path
import pandas as pd
from pyadlml.constants import VALUE, TIME, DEVICE, END_TIME, START_TIME, ACTIVITY
from pyadlml.dataset._core.obj import Data, DataFetcher, MegaDownloader
from pyadlml.dataset.io import fetcher, correct_acts_and_devs, download_from_mega

ARAS_URL = 'https://mega.nz/file/hVpRADoZ#GLLZDV4Y-vgdEeEDTXnFxeG3eKllhTljMM1RK-eGyh4'
ARAS_FILENAME = 'aras.zip'

RES_1 = 'resident1'
RES_2 = 'resident2'

__all__ = ['fetch_aras']

class ArasFetcher(DataFetcher):
    def __init__(self, *args, **kwargs):

        downloader = MegaDownloader(
            url=ARAS_URL,
            fn=ARAS_FILENAME,
            url_cleaned=None,
            fn_cleaned=None,
        )

        super().__init__(
            dataset_name='aras',
            downloader=downloader,
            # TODO change to true
            correct_activities=False,
            correct_devices=False
        )

    def load_data(self, folder_path: Path) -> dict:

        device_map = _get_device_map(folder_path)
        activity_map = _get_activity_map(folder_path)
        df = _read_data(folder_path, activity_map, device_map)

        df_res1_act = _create_activity_df(df, RES_1)
        df_res2_act = _create_activity_df(df, RES_2)

        df_dev = _create_device_df(df)
        lst_dev = df_dev[DEVICE].unique()
        from pyadlml.dataset.plotly.util import ActivityDict
        act_dict = ActivityDict({
                f'df_activities_{RES_1}':df_res1_act,
                f'df_activities_{RES_2}':df_res2_act,
        })
        return dict(
            df_devices=df_dev,
            device_list=lst_dev,
            df_activities=act_dict,
            activity_list={
                f'activity_list_{RES_1}': df_res1_act[ACTIVITY].unique(),
                f'activity_list_{RES_2}': df_res2_act[ACTIVITY].unique(),
            }
        )

def fetch_aras(keep_original=True, cache=True, retain_corrections=False,
               folder_path=None):
    """
    Fetches the aras dataset from the internet. The original dataset or its cached version
    is stored in the :ref:`data home <storage>` folder.

    Parameters
    ----------
    keep_original : bool, default=True
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool, default=True
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, default=False
        When set to *true* data points that are changed or dropped during preprocessing
        are listed in the respective attributes of the data object.  Fore more information
        about the attributes refer to the :ref:`user guide <error_correction>`.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_aras
    >>> data = fetch_aras()
    >>> dir(data)
    >>> [..., df_activities, df_devices, ...]

    Returns
    -------
    data : object
    """
    return ArasFetcher()(keep_original=keep_original, cache=cache, #load_cleaned=load_cleaned,
                              retain_corrections=retain_corrections, folder_path=folder_path)


    return data


def _get_device_map(data_path):
    """
    Parameters
    ----------
    data_path : string
        path to folder where aras dataset resides
    Returns
    -------
    devices : pd.DataFrame
        dataframe with information about devices 
        the index represents the column of the sensor table
    """
    cols = ['Sensor ID', 'Sensor Type', 'Place']
    devices = pd.DataFrame(columns=cols)
    with open(os.path.join(data_path, 'README'), 'r') as file:
        lines = file.readlines()
        # create header
        for line in lines[7:27]:
            s = list(filter(None, line.split('\t')))
            s[-1:] = [s[-1:][0][:-1]]
            tmp = dict(zip(cols, s[1:]))
            devices = devices.append(tmp, ignore_index=True)

        file.close()
    return devices


def _get_activity_map(data_path):
    """
    Parameters
    ----------
    data_path : string
        path to folder where aras dataset resides
    Returns
    -------
    devices : pd.DataFrame
        Contains id matching label of activity 
    """    
    cols = ['ID', 'activity']
    activities = pd.DataFrame(columns=cols)

    with open(os.path.join(data_path, 'README'), 'r') as file:
        lines = file.readlines()
        # create header
        for line in lines[33:60]:
            s = list(filter(None, line.split('\t')))
            s[-1:] = [s[-1:][0][:-1]]
            tmp = dict(zip(cols, s))
            activities = activities.append(tmp, ignore_index=True)

        file.close()
    activities['ID'] = activities['ID'].astype(int)
    return activities

def _read_data(data_path, activity_map, device_map):
    #DEF
    file_name = 'DAY_{}.txt'
    cols = list(device_map['Sensor ID']) + ([RES_1, RES_2])
    res = pd.DataFrame(columns = cols)

    # read in all files
    for i in range(1,31):
        file_path = os.path.join(data_path, file_name.format(i))
        df = pd.read_table(file_path, sep=' ', header=None)
        df.columns = cols
        if i < 10:
            date = '2000-01-0{}'.format(i)
        else:
            date = '2000-01-{}'.format(i)
        df.index = pd.date_range(date, periods=86400, freq='s')

        # delete all rows where nothing changes
        mask = (df.diff(axis=0).sum(axis=1) == 0.0)
        mask.iloc[0] = False
        df = df[~mask]

        # append all dataframes
        res = pd.concat([res, df])

    # label activities
    act_dict = dict(zip(activity_map['ID'],activity_map['activity']))
    res[RES_1] = res[RES_1].map(act_dict)
    res[RES_2] = res[RES_2].map(act_dict)
    return res


def _create_activity_df(df, res_name):
    # bring data into representation 3 
    df = pd.DataFrame(df[res_name]).reset_index()
    
    mask_start = df[res_name] != df[res_name].shift(1)
    df_start = df[mask_start]
    df_start = df_start.rename(columns={'index' : START_TIME})
    df_start = df_start.reset_index(drop=True)

    mask_end = df[res_name] != df[res_name].shift(-1)
    df_end = df[mask_end].reset_index(drop=True)
    
    df_start[END_TIME] = df_end['index']
    
    df = df_start[[START_TIME, END_TIME, res_name]]
    df = df.sort_values(by=START_TIME)
    df = df.rename(columns={res_name: ACTIVITY})
    df = df[df[ACTIVITY] != 'Other'] # remove Other activity as it is the same concept as 'idle'
    return df


def _create_device_df(df):
    """ gets a raw representation and returns devices in rep1
    Parameters
    ----------
    df : pd.DataFrame
        raw representation 
    """
    df = df.copy().iloc[:, :-2] # exclude activities
    mask = (df.diff(axis=0).sum(axis=1) == 0.0)
    mask.iloc[0] = False
    df = df[~mask]
    
    # is true where the devices change
    mask_0to1 = (df.diff(axis=0) == 1)
    mask_1to0 = (df.diff(axis=0) == -1)
    
    # for every device append the rows where the device changes state
    res = pd.DataFrame(columns=[TIME, DEVICE, VALUE])
    for device in df.columns:
        dev_0to1 = pd.DataFrame(df[device][mask_0to1[device]])
        dev_1to0 = pd.DataFrame(df[device][mask_1to0[device]])
        tmp = pd.concat([dev_0to1, dev_1to0]).reset_index()
        tmp.columns = [TIME, VALUE]
        tmp[DEVICE] = device
        
        res = pd.concat([res, tmp])
    
    res = res.sort_values(by=TIME).reset_index(drop=True)
    res[VALUE] = res[VALUE].astype(bool)
    return res




import pandas as pd

from pyadlml.constants import ACTIVITY, DEVICE, END_TIME, START_TIME
from pyadlml.dataset._core.activities import correct_activities
from pyadlml.dataset._core.devices import device_boolean_on_states_to_events
from pyadlml.dataset._core.obj import Data
from pyadlml.dataset.io import fetcher, correct_devs, download_from_mega
import os

UCI_ADL_BINARY_URL = 'https://mega.nz/file/AQIgDQJD#oximAQFjexTKwNP3WYzlPnOGew06YSQ2ef85vvWGN94'
UCI_ADL_BINARY_FILENAME = 'uci_adl_binary.zip'
ORD_A = 'OrdonezA'
ORD_B = 'OrdonezB'


@correct_devs   # only correct devices as activities are explicitly handled in function body
@fetcher('uci_adl_binary', download_from_mega, UCI_ADL_BINARY_FILENAME, UCI_ADL_BINARY_URL)
def fetch_uci_adl_binary(subject='OrdonezA', keep_original=True, cache=True,
                         retain_corrections=False, folder_path=None) -> Data:
    """

    Parameters
    ----------
    subject : str of {'OrdonezA', 'OrdonezB'}, default='OrdonezA'
        decides which dataset of the two houses is loaded.
    keep_original : bool, default=True
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool
        When set to *true* data points that are changed or dropped during preprocessing
        are listed in the respective attributes of the data object.  Fore more information
        about the attributes refer to the :ref:`user guide <error_correction>`.


    Returns
    -------
    data : object
    """
    assert subject in [ORD_A, ORD_B]

    # fix path and Ordonez B activity file
    sub_dev_file = os.path.join(folder_path, '{}_Sensors.txt'.format(subject))
    if subject == ORD_B:
        fix_OrdonezB_ADLS(os.path.join(folder_path, 'OrdonezB_ADLs.txt'))
        sub_act_file = os.path.join(folder_path, '{}_ADLs_corr.txt'.format(subject))
    else:
        sub_act_file = os.path.join(folder_path, '{}_ADLs.txt'.format(subject))

    # load activities and devices
    df_act = _load_activities(sub_act_file)
    df_dev, df_loc = _load_devices(sub_dev_file)
    df_dev = device_boolean_on_states_to_events(df_dev)

    lst_act = df_act[ACTIVITY].unique()
    lst_dev = df_dev[DEVICE].unique()

    data = Data(None, df_dev, activity_list=lst_act, device_list=lst_dev)
    data.df_dev_rooms = df_loc

    # manual activity correction for OrdonezB
    if subject == ORD_B:
        # the activity grooming is often overlapped by sleeping
        # Since grooming is more important make it the dominant activity (opinionated)
        df_act, correction_act = correct_activities(df_act, excepts=['Grooming'], retain_corrections=retain_corrections)

    elif subject == ORD_A:
        # Manually replace 3 rows where START_TIME > END_TIME in the activity files
        # 72 "2011-12-01 19:28:51  2011-12-01 16:29:59  Toileting"
        # to "2011-12-01 19:28:51  2011-19-01 16:29:59  Toileting"
        # 81 "2011-12-02 12:20:41  2011-12-01 10:20:59  Grooming"
        # to "2011-12-02 12:20:41  2011-12-01 12:20:59  Grooming"
        # 83 "2011-12-02 12:27:47  2011-12-01 11:35:49  Breakfast"
        # to "2011-12-02 12:27:47  2011-12-01 12:35:49  Breakfast"
        df_act.iat[69, 1] = pd.Timestamp('2011-12-01 19:29:59')
        df_act.iat[78, 1] = pd.Timestamp('2011-12-02 12:20:59')
        df_act.iat[80, 1] = pd.Timestamp('2011-12-02 12:35:49')

        df_act, correction_act = correct_activities(df_act, retain_corrections=retain_corrections)

    data.set_activity_df(df_act, subject)

    if retain_corrections:
        data.set_activity_correction(correction_act, subject)

    return data


def fix_OrdonezB_ADLS(path_to_file):
    """ fixes inconsistent use of tabs for delimiter in the file
    Parameters
    ----------
    path_to_file : str
        path to the file OrdonezB_ADLs.csv
    """
    
    path_corrected = path_to_file[:-17] + 'OrdonezB_ADLs_corr.txt'
    
    with open(path_to_file, 'r') as f_o, open(path_corrected, 'w') as f_t:
        for i, line in enumerate(f_o.readlines()):            
            if i in [0,1]: 
                f_t.write(line)  
                continue
            s = line.split()
            assert len(s) == 5
            new_line = s[0]+' '+s[1]+'\t\t'+s[2]+' '+s[3]+'\t\t'+s[4]                        
            f_t.write(new_line + "\n")
        f_t.close()
        f_o.close()

def _load_activities(act_path):
    df_act = pd.read_csv(act_path, delimiter='\t+', skiprows=[0,1], 
                         names=[START_TIME, END_TIME, ACTIVITY], engine='python')
    df_act[START_TIME] = pd.to_datetime(df_act[START_TIME])
    df_act[END_TIME] = pd.to_datetime(df_act[END_TIME])
    return df_act

def _load_devices(dev_path):
    df_dev = pd.read_csv(dev_path, delimiter='\t+', skiprows=[0, 1], 
                         names=[START_TIME, END_TIME, 'Location', 'Type', 'Place'], 
                         engine='python')
    df_dev[DEVICE] = df_dev['Place'] + ' ' + df_dev['Location'] + ' ' + df_dev['Type']
    
    # get room mapping devices
    df_locs = df_dev.copy().groupby([DEVICE, 'Type', 'Place', 'Location']).sum()
    df_locs = df_locs.reset_index().drop([START_TIME, END_TIME], axis=1)

    df_dev = df_dev[[START_TIME, END_TIME, DEVICE]]
    df_dev[START_TIME] = pd.to_datetime(df_dev[START_TIME])
    df_dev[END_TIME] = pd.to_datetime(df_dev[END_TIME])
    return df_dev, df_locs


from pathlib import Path
import os

from pyadlml.dataset.io import DATA_HOME_FOLDER_NAME, set_data_home
import pyadlml.dataset as dataset
from pyadlml.constants import PRIMARY_COLOR, SECONDARY_COLOR, CM_SEQ_MAP, CM_DIV_MAP
from pyadlml.util import (
    ENV_PARALLEL,
    set_primary_color,
    set_diverging_color,
    set_secondary_color,
    set_sequential_color
)

# set data home folder to the current working directory
path = [str(Path('/tmp/').joinpath(DATA_HOME_FOLDER_NAME))]
DATA_HOME = path
set_data_home(path[0])  # do this in order to create the folder

# default for parallel execution
os.environ[ENV_PARALLEL] = str(False)

set_primary_color(PRIMARY_COLOR)
set_secondary_color(SECONDARY_COLOR)
set_diverging_color(CM_DIV_MAP)
set_sequential_color(CM_SEQ_MAP)



from pyadlml.dataset.stats.discrete import contingency_table_01, cross_correlation
from pyadlml.dataset.plot.util import heatmap_contingency as hm_cont
from pyadlml.dataset.plot.util import annotate_heatmap, heatmap, heatmap_square, \
    _num_bars_2_figsize, func_formatter_log_1
from pyadlml.util import get_primary_color, get_diverging_color
import matplotlib.pyplot as plt
import matplotlib
import pandas as pd 
import numpy as np

def heatmap_contingency(X, y, rep='', z_scale=None, figsize=None, numbers=True):
    """ plots the contingency between features and labels of data
    Parameters
    ----------
    X: pd.DataFrame
        one of the representation raw, lastfired, changepoint
    y: pd.DataFrame
        a series of labels
    rep: string
        the name of the representation to add to the title
    """
    title = rep + " On/Off contingency"
    cbarlabel = 'counts'
    valfmt = ("{x:.0f}" if z_scale!='log' else func_formatter_log_1)

    df_con = contingency_table_01(X, y)
    vals = df_con.values.T
    acts = df_con.columns.values
    devs = list(df_con.index)
    
    # format labels by replacing every 'on'
    for i, dev in enumerate(devs):
        if 'On' in dev:
            devs[i] = 'On'
    return hm_cont(acts, devs, vals, title, cbarlabel, z_scale=z_scale, figsize=figsize, valfmt=valfmt, numbers=numbers)

def hist_activities(y, scale=None, color=None, figsize=(9,3)):
    """
    Parameters
    ----------
    y: np array
        label of strings
    scale: None or log
    """
    title = 'Label occurence'
    xlabel = 'counts'
    color = (get_primary_color() if color is None else color)

    ser = pd.Series(data=y).value_counts()
    ser = ser.sort_values(ascending=True)

    figsize = (_num_bars_2_figsize(len(ser)) if figsize is None else figsize)
    fig, ax = plt.subplots(1, 1, figsize=figsize)

    if scale == 'log': 
        plt.xscale('log')

    ax.barh(ser.index, ser.values, orientation='horizontal', color=color)
    plt.xlabel(xlabel)
    fig.suptitle(title)


def corr_devices_01(rep, figsize=(19,14)):
    """ correlation between the on and off states of every device.
    Parameters
    ----------
    rep: pd.DataFrame
        A dataframe where columns are devices and rows are a binary representation
        e.g raw, changepoint, lastfired representation
    
    """
    df = rep.iloc[:,:-1] # get raw without activities
    df = df.reset_index(drop=True)
    df = df.astype(int)

    for device in df.columns:
        mask1 = df[[device]] == 1
        col_off = device + ' Off'
        df[[col_off]] = df[[device]].astype(bool)
        df[[col_off]] = ~df[[col_off]]
        df[[col_off]] = df[[col_off]].astype(int)
        df = df.rename(columns={device: device + ' On'})
        
    dev_count = int(len(df.columns)/2)
    
    ct = df.corr()
    ct_on_on = ct.iloc[:dev_count, :dev_count]
    ct_on_on_values = ct_on_on.values.T
    ct_on_on_devs = list(ct_on_on.index)
    
    ct_on_off = ct.iloc[dev_count:, :dev_count]
    ct_on_off_values = ct_on_off.values.T
    ct_on_off_rows = list(ct_on_off.index)
    ct_on_off_cols = list(ct_on_off.columns)

    fig, (ax1, ax2) = plt.subplots(1,2, figsize=figsize)
    
    # plot on-on corr
    im, cbar = heatmap(ct_on_on_values, ct_on_on_devs, ct_on_on_devs, 
                       ax=ax1, cmap='PuOr', cbarlabel='counts',
                       vmin=-1, vmax=1)

    texts = annotate_heatmap(im, textcolors=("black", "white"), valfmt="{x:.2f}")
    
    # plot on-off corr
    im, cbar = heatmap(ct_on_off_values, ct_on_off_rows, ct_on_off_cols, 
                       ax=ax2, cmap='PuOr', cbarlabel='counts',
                       vmin=-1, vmax=1)
    
    texts = annotate_heatmap(im, textcolors=("black", "white"), valfmt="{x:.2f}")
    
    ax1.set_title("Correlation of on-on signals")
    ax2.set_title("Correlation of on-off signals")
    fig.tight_layout()
    plt.show()

def heatmap_cross_correlation(df_dev, figsize=(10,8)):


    title = 'Devices cross-correlation'
    cmap = get_diverging_color()
    cbarlabel = 'similarity'

    ct = cross_correlation(df_dev)
    vals = ct.values.T.astype(np.float64)
    devs = list(ct.index)
    
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(vals, devs, devs, ax=ax, cmap=cmap, cbarlabel=cbarlabel,
                       vmin=-1, vmax=1)
    annotate_heatmap(im, textcolors=("black", "white"), valfmt="{x:.2f}", threshold=0.6)
    ax.set_title(title)
    fig.tight_layout()
    plt.show()


import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd



def _create_cmap(self, cmap, n=20):
    import matplotlib.colors as colors
    assert n <= 20
    minval = 0
    maxval = n
    #lst = cmap(np.linspace(minval, maxval, n))
    #formt = 'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval)
    lst = np.arange(minval, maxval)
    new_colors = cmap(lst)
    new_cmap = colors.ListedColormap(new_colors)
    #new_cmap = colors.LinearSegmentedColormap.from_list(
    #    formt,
    #    lst)
    return new_cmap

def plot_duration_distributions(cls, true_z, true_x, hmm, hsmm, hm_z, state_sel, img_file_path):

    from scipy.stats import nbinom
    # Plot the true and inferred duration distributions
    """
    N = the number of infered states
        how often the state was inferred
        blue bar is how often when one was in that state it endured x long
    x = maximal duration in a state


    red binomial plot
        for the hmm it is 1 trial and the self transitioning probability
        for the hsmm it is

    Negativ binomial distribution for state durations
    
        NB(r,p)
            r int, r>0
            p = [0,1] always .5 wk des eintretens von erfolgreicher transition
            r = anzahl erflogreiche selbst transitionen  befor man etwas anderes (trans in anderen
            zustand sieht)
    """
    from ssm.util import rle

    hmm_z = hmm.most_likely_states(true_x)
    hsmm_z = hsmm.most_likely_states(true_x)

    K = hmm.K
    true_states = true_z
    true_states, true_durations = rle(true_states)
    hmm_inf_states, hmm_inf_durations = rle(hmm_z)
    hsmm_inf_states, hsmm_inf_durations = rle(hsmm_z)

    max_duration = max(np.max(true_durations), np.max(hsmm_inf_durations), np.max(hmm_inf_durations))
    max_duration = 50
    dd = np.arange(max_duration, step=1)

    n_cols = len(state_sel)
    n_rows = 3

    height = 9
    width = 3*n_cols
    plt.figure(figsize=(width, height))
    legend_label_hmm = 'hmm'
    legend_label_hsmm = 'hsmm'
    #for k in range(K):
    #n_cols = K
    for col, act in enumerate(state_sel):
        # Plot the durations of the true states
        index = col+1
        plt.subplot(n_rows, n_cols, index)
        """
        get the durations where it was gone into the state k =1
        state_seq: [0,1,2,3,1,1]
        dur_seq: [1,4,5,2,4,2]
            meaning one ts in state 0, than 4 in state 1, 5 in state 2, so on and so forth
        x = [4,4,2]
        """
        enc_state = hm_z[act]
        x = true_durations[true_states == enc_state] - 1
        plt.hist(x, dd, density=True)
        #n = true_hsmm.transitions.rs[k]
        #p = 1 - true_hsmm.transitions.ps[k]
        #plt.plot(dd, nbinom.pmf(dd, n, p),
        #         '-k', lw=2, label='true')
        #if k == K - 1:
        #    plt.legend(loc="lower right")
        plt.title("{} (N={})".format(act, np.sum(true_states == enc_state)))

        # Plot the durations of the inferred states of hmm
        index = 2*n_cols+col+1
        plt.subplot(n_rows, n_cols, index)
        plt.hist(hmm_inf_durations[hmm_inf_states == enc_state] - 1, dd, density=True)
        plt.plot(dd, nbinom.pmf(dd, 1, 1 - hmm.transitions.transition_matrix[enc_state, enc_state]),
                 '-r', lw=2, label=legend_label_hmm)
        if col == n_cols - 1:
            plt.legend(loc="lower right")
        #plt.title("State {} (N={})".format(k+1, np.sum(hmm_inf_states == k)))
        plt.title("{} (N={})".format(act, np.sum(hmm_inf_states == enc_state)))

        # Plot the durations of the inferred states of hsmm
        index = n_cols+col+1
        plt.subplot(n_rows, n_cols, index)
        plt.hist(hsmm_inf_durations[hsmm_inf_states == enc_state] - 1, dd, density=True)
        plt.plot(dd, nbinom.pmf(dd, hsmm.transitions.rs[enc_state], 1 - hsmm.transitions.ps[enc_state]),
                 '-r', lw=2, label=legend_label_hsmm)
        if col == n_cols - 1:
            plt.legend(loc="lower right")
        plt.title("{} (N={})".format(act, np.sum(hsmm_inf_states == enc_state)))

    plt.tight_layout()
    #plt.show()
    plt.savefig(img_file_path)
    plt.clf()


def save_plot_feature_importance(self, fp):
    """
    Parameters
    ----------
    fp : str
        the file path to save fp to
    """
    self._feature_importance.save_plot_feature_importance(fp)


def create_skater_stuff(self, dataset):
    test_z, test_x = dataset.get_all_labeled_data() # states, obs
    self._feature_importance = FeatureImportance(self._model, test_x, test_z)

    #from  hassbrain_algorithm.benchmark.explanation import create_explanator
    #self._explanator = self._create_explanator(test_x, test_z)


def get_feature_importance(self):
    # is the amount of samples the feature takes
    N_SAMPLES = 18
    imp_series = self._skater_interpreter.feature_importance.feature_importance(
        self._skater_model,
        n_samples=N_SAMPLES,
        ascending=True,
        progressbar=False,
        # model-scoring: difference in log_loss or MAE of training_labels
        # given perturbations. Note this vary rarely makes any significant
        # differences
        method='model-scoring')
    # corss entropy or f1 ('f1', 'cross_entropy')
    #scorer_type='cross_entropy') # type: Figure, axes
    #scorer_type='f1') # type: Figure, axes
    return imp_series

def plot_feature_importance(self):
    fig, ax = self._skater_interpreter.feature_importance.plot_feature_importance(
        self._skater_model,
        ascending=True,
        ax=None,
        progressbar=False,
        # model-scoring: difference in log_loss or MAE of training_labels
        # given perturbations. Note this vary rarely makes any significant
        # differences
        method='model-scoring')
    # corss entropy or f1 ('f1', 'cross_entropy')
    #scorer_type='cross_entropy') # type: Figure, axes
    #scorer_type='f1') # type: Figure, axes
    fig.show()



import sys
import pathlib

from pyadlml.dataset.util import remove_days

working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest
import pandas as pd

from pyadlml.dataset import load_act_assist
from pyadlml.dataset import ACTIVITY, DEVICE, END_TIME, START_TIME, VAL, TIME
from pyadlml.feature_extraction import extract_time_bins, extract_day_of_week

SUBJECT_ADMIN_NAME = 'admin'

def timestamp(string):
    return pd.to_datetime(string, dayfirst=True)

def df(d):
    return pd.DataFrame(d, columns=[START_TIME, END_TIME, ACTIVITY])


class TestDirtyDataset(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/dirty_dataset'
        self.data = load_act_assist(dataset_dir, subjects=[SUBJECT_ADMIN_NAME])

        self.days_to_remove = ['02.01.2000']
        self.offsets = ['3h']
        self.data_case_1 = [
            [timestamp('01.01.2000 12:00:00'), timestamp('02.01.2000 12:00:00'), 'act_1'],
            [timestamp('03.01.2000 08:00:00'), timestamp('03.01.2000 10:00:00'), 'act_2'],
            [timestamp('03.01.2000 10:00:01'), timestamp('03.01.2000 18:00:00'), 'act_3'],
        ]
        self.data_case_2 = [
            [timestamp('01.01.2000 12:00:00'), timestamp('02.01.2000 12:00:00'), 'act_1'],
            [timestamp('03.01.2000 06:00:00'), timestamp('03.01.2000 18:00:00'), 'act_2'],
        ]
        self.data_case_3 = [
            [timestamp('01.01.2000 06:00:00'), timestamp('01.01.2000 20:00:00'), 'act_1'],
            [timestamp('01.01.2000 20:00:01'), timestamp('01.01.2000 23:00:00'), 'act_2'],
            [timestamp('02.01.2000 18:00:00'), timestamp('03.01.2000 12:00:00'), 'act_2'],
        ]
        self.data_case_4 = [
            [timestamp('01.01.2000 18:00:00'), timestamp('01.01.2000 23:00:00'), 'act_1'],
            [timestamp('02.01.2000 20:00:00'), timestamp('03.01.2000 12:00:00'), 'act_2'],
        ]

    def test_extract_time_bins(self):
        from pyadlml.dataset.plot.activities import correction

        df_activities = df(self.data_case_1)
        print('case 1: \n', df_activities)
        _, df_activities_2 = remove_days(self.df_devices, df_activities, self.days_to_remove)
        print(df_activities_2, '\n', '~'*100)
        #correction(df_activities, df_activities_2).show()
        df_activities = df(self.data_case_2)
        print('case 2: \n', df_activities)
        _, df_activities_2 = remove_days(df_devices, df_activities, self.days_to_remove)
        print(df_activities_2, '\n', '~'*100)
        #correction(df_activities, df_activities_2).show()
        df_activities = df(self.data_case_3)
        print('case 3: \n', df_activities)
        _, df_activities_2 = remove_days(df_devices, df_activities, self.days_to_remove)
        print(df_activities, '\n', '~'*100)
        correction(df_activities, df_activities_2).show()
        df_activities = df(self.data_case_4)
        print('case 4: \n', df_activities)
        _, df_activities_2 = remove_days(df_devices, df_activities, self.days_to_remove)
        print(df_activities, '\n', '~'*100)
        correction(df_activities, df_activities_2).show()





if __name__ == '__main__':
    unittest.main()


from pyadlml.dataset import fetch_kasteren_2010, set_data_home, \
    TIME, START_TIME, END_TIME, ACTIVITY, DEVICE
from pyadlml.dataset.stats.activities import coverage
from pyadlml.dataset.util import infer_dtypes, fetch_data_by_string, DATASET_STRINGS
import argparse

path = '/media/data/ml_datasets/pyadlml'


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset', type=str, default='amsterdam',
                        choices=DATASET_STRINGS,
                        help='Select an avaliable dataset.'
                        )
    args = parser.parse_args()

    # Setup data
    set_data_home(path)
    data = fetch_data_by_string(args.dataset)

    df_acts = data.df_activities
    df_devs = data.df_devices

    # Determine start and end_time
    start_time = min(df_devs[TIME].iloc[0], df_acts[START_TIME].iloc[0])
    end_time = max(df_devs[TIME].iloc[-1], df_acts[END_TIME].iloc[-1])

    # Compute activity statistics
    nr_activities = len(df_acts[ACTIVITY].unique())
    nr_activity_recordings = len(df_acts)

    # Compute device statistics
    nr_devices = len(df_devs[DEVICE].unique())
    nr_device_recordings = len(df_devs)
    dev_types = [k for k, v in infer_dtypes(df_devs).items() if len(v) > 0]


    act_cov_dp = coverage(df_acts, df_devs, datapoints=True)
    act_cov_time = coverage(df_acts, df_devs, datapoints=False)

    print(f'From:\t\t{start_time}')
    print(f'To:\t\t\t{end_time}')
    print(f'Activites:\t{nr_activities}/{nr_activity_recordings}')
    print(f'Devices:\t{nr_devices}/{nr_device_recordings}')
    print(f'DeviceType:\t{dev_types}')
    print(f'Coverage: \t{act_cov_time:.2f}/{act_cov_dp:.2f}')
